"""
Created on May 4, 2015

@author: Derek Wood
"""

import logging
import os
import re
import shutil
import subprocess
import sys
import tempfile
from typing import Union

from bi_etl.informatica.exceptions import NoObjects
from bi_etl.informatica.pm_config import PMCMDConfig
from bi_etl.utility import dict_to_str, line_counter


class PMREP(object):
    SETTINGS_SECTION = 'INFORMATICA'

    def __init__(self, config: PMCMDConfig):
        self.config = config
        self.log = logging.getLogger(f"{self.__class__.__module__}.{self.__class__.__name__}")
        self.f_dev_null = open(os.devnull, 'w')
        self.control_file_name = "Control_import_No_folder_rep_change.xml"
        self._folder = None

    def setup_inf_path(self):
        user_dir = os.path.expanduser('~')
        os.environ['PATH'] = ':'.join([os.path.join(user_dir, 'bin'),
                                       self.informatica_bin_dir(),
                                       '/usr/bin',
                                       ]
                                      )
        os.environ['LD_LIBRARY_PATH'] = self.informatica_bin_dir()

    def infa_home(self) -> Union[str, None]:
        if 'INFA_HOME' in os.environ:
            return os.environ['INFA_HOME']
        else:
            return self.config.INFA_HOME

    def informatica_bin_dir(self):
        return os.path.join(self.infa_home(), 'server', 'bin')

    def informatica_pmcmd(self):
        return os.path.join(self.informatica_bin_dir(), 'pmcmd')

    def usersecuritydomain(self):
        return self.config.USER_SECURITY_DOMAIN

    def user_id(self):
        return self.config.user_id

    def password(self):
        return self.config.get_password()

    def set_password_in_env(self):
        os.environ['INFA_PM_PASSWORD'] = self.password()

    def repository(self):
        return self.config.REPOSITORY

    def service(self):
        return self.config.SERVICE

    def domain(self):
        return self.config.DOMAIN

    def folder(self):
        if self._folder is None:
            self._folder = self.config.DEFAULT_FOLDER
        return self._folder

    def informatica_pmrep(self):
        return os.path.join(self.informatica_bin_dir(), 'pmrep')

    def connect(self):
        pmrep_cmd = [self.informatica_pmrep(),
                     'connect',
                     '-r', self.repository(),
                     '-d', self.domain(),
                     '-n', self.user_id(),
                     '-X', 'INFA_PM_PASSWORD'
                     ]
        self.set_password_in_env()
        self.setup_inf_path()

        self.log.info("pmrep Connecting to Informatica")
        try:
            if self.log.getEffectiveLevel() >= logging.DEBUG:
                file_out = sys.stdout
            else:
                file_out = self.f_dev_null
            subprocess.check_call(pmrep_cmd, stdout=file_out)
        except subprocess.CalledProcessError as e:
            self.log.error("Error code " + str(e.returncode))
            self.log.error("From " + ' '.join(e.cmd))
            self.log.error(e.output)
            raise e

    def cleanup(self):
        pmrep_cmd = [self.informatica_pmrep(), 'cleanup']
        try:
            messages = subprocess.check_output(pmrep_cmd, stderr=subprocess.STDOUT)
            self.log.debug(messages)
        except subprocess.CalledProcessError as e:
            self.log.error("Error code " + str(e.returncode))
            self.log.error("From " + ' '.join(e.cmd))
            self.log.error(e.output)
        finally:
            self.f_dev_null.close()

    def get_objects(self, object_type, folder_name):
        # print "pmrep ListObjects -o " + objectType + ' -f ' + folderName
        object_list = list()
        process = subprocess.Popen(
            [
                self.informatica_pmrep(),
                'ListObjects',
                '-o', object_type,
                '-f', folder_name
            ],
            stdout=subprocess.PIPE
        )
        count = 0
        found_invoked = False
        found_blank_line = False
        for line in iter(process.stdout.readline, ''):
            # End on line .ListObjects completed successfully.
            if line.startswith(b'.ListObjects'):
                break
            if found_blank_line:
                count += 1
                # if count >= 15: break
                parts = line.rstrip(b'\n').split(b' ')
                subtype = parts[0]
                if len(parts) == 2:
                    reusable = 'reusable'
                    name = parts[1]
                else:
                    reusable = parts[1]
                    name = parts[2]
                # print "parts = " + pformat(parts)
                if reusable == 'reusable':
                    # print "subtype = " + subtype + " name = " + name
                    object_dict = {'objectType': object_type,
                                   'subtype': subtype,
                                   'name': name,
                                   'folderName': folder_name
                                   }
                    object_list.append(object_dict)
            if line.startswith(b'Invoked'):
                found_invoked = True
            if found_invoked and line == '\n':
                found_blank_line = True
        return object_list

    def get_objects_from_query(self, query_name):
        # pmrep  executequery -q $INFA_QUERY_NAME -t shared -u ${OUTPUT_PATH}\${INFA_QUERY_NAME}_results.txt
        tempDir = tempfile.mkdtemp()
        temp_file = os.path.join(tempDir, 'query.out')
        obj_list = list()
        try:
            if self.log.getEffectiveLevel() >= logging.DEBUG:
                file_out = sys.stdout
            else:
                file_out = self.f_dev_null
            subprocess.check_call([self.informatica_pmrep(),
                                   'executequery',
                                   '-q', query_name,
                                   '-u', temp_file
                                   ],
                                  stdout=file_out)
            if os.path.exists(temp_file):
                count = 0
                with open(temp_file, 'r') as f:
                    for line in f:
                        count += 1
                        # if count >= 15: break
                        parts = line.rstrip('\n').split(',')
                        folder = parts[1]
                        name = parts[2]
                        object_type = parts[3]
                        subtype = parts[4]
                        # version = parts[5]
                        if len(parts) == 7:
                            reusable = parts[6]
                        else:
                            reusable = 'reusable'
                        if reusable == 'reusable':
                            # print("parts {} = 0-excluded {}" .format(len(parts),[parts[i] for i in range(1,7)]))
                            object_dict = {'objectType': object_type,
                                           'subtype': subtype,
                                           'name': name,
                                           'folder': folder
                                           }
                            obj_list.append(object_dict)
            else:  # query.out not created
                pass
        except subprocess.CalledProcessError:
            raise RuntimeError("Error executing query {name}".format(name=query_name))
        finally:
            # Cleanup temp
            shutil.rmtree(tempDir)
        return obj_list

    def deleteObject(self, objectDict):
        # pmrep  DeleteObject -o <object_type> -f <folder_name> -n <object_name>
        pmrep_cmd = [self.informatica_pmrep(),
                     'DeleteObject',
                     '-f', objectDict['folder'],
                     '-o', objectDict['type'],
                     '-n', objectDict['name']
                     ]

        # Include subtype if required
        if objectDict['type'].lower() in ('task', 'transformation'):
            pmrep_cmd.append('-t')
            pmrep_cmd.append(objectDict['subtype'])

        try:
            if self.log.getEffectiveLevel() >= logging.DEBUG:
                file_out = sys.stdout
            else:
                file_out = self.f_dev_null
            subprocess.check_call(pmrep_cmd, stdout=file_out)
        except subprocess.CalledProcessError as e:
            self.log.error("Error code " + str(e.returncode))
            self.log.error("From " + ' '.join(e.cmd))
            self.log.error(e.output)

    def exportObject(self, objectDict, dependents, outputPath):
        # pmrep  objectexport -f $FOLDER -n "$NAME" -o "$TYPE" -t "$SUBTYPE" $DEPENDENTS_OPTIONS -u "${TYPE}s/${NAME}.xml"
        pmrep_cmd = [
            self.informatica_pmrep(),
            'objectexport',
            '-f', objectDict['folder'],
            '-n', objectDict['name'],
            '-o', objectDict['type']
        ]

        # Include subtype if required
        if objectDict['type'].lower() in ('task', 'transformation'):
            pmrep_cmd.append('-t')
            pmrep_cmd.append(objectDict['subtype'])

        # include all dependents or only non-reusable dependents
        if dependents:
            pmrep_cmd.append('-m')  # [-m (export pk-fk dependency)]
            pmrep_cmd.append('-s')  # [-s (export objects referred by shortcut)]
            pmrep_cmd.append('-b')  # [-b (export non-reusable dependents)]
            pmrep_cmd.append('-r')  # [-r (export reusable dependents)]
        else:
            pmrep_cmd.append('-b')  # [-b (export non-reusable dependents)]

        pmrep_cmd.append('-u')
        pmrep_cmd.append(outputPath)

        try:
            messages = subprocess.check_output(pmrep_cmd, stderr=subprocess.STDOUT)
            count_xml_lines = line_counter.bufcount(outputPath)
            errors = re.findall('^.*<Warning>.*$|^.*<Error>.*$', messages, re.MULTILINE)
            if len(errors) > 0:
                self.log.error(errors)
                # noinspection PyTypeChecker
                return '\n'.join(errors)
            elif count_xml_lines <= 3:
                print("WARNING: No valid objects exported")
                os.remove(outputPath)
                raise NoObjects()
        except subprocess.CalledProcessError as e:
            messages = e.output
            print("Error code " + str(e.returncode))
            print("From " + ' '.join(e.cmd))
            print(messages)
            return messages

    def validateObject(self, objectDict):
        #
        # pmrep validate {{-n <object_name>  -o <object_type (mapplet, mapping, session, worklet, workflow)>
        #              [-v <version_number>] [-f <folder_name>]} |  -i <persistent_input_file>}
        #              [-s (save upon valid) [-k (check in upon valid) [-m <check_in_comments>]]]
        #              [-p <output_option_types (valid, saved, skipped, save_failed, invalid_before, invalid_after, or all)>
        #              [-u <persistent_output_file_name>]  [-a (append)]
        #              [-c <column_separator>] [-r <end-of-record_separator>] [-l <end-of-listing_indicator>] [-b (verbose)]
        #
        pmrep_cmd = [self.informatica_pmrep(),
                     'validate',
                     '-f', objectDict['folder'],
                     '-n', objectDict['name'],
                     '-o', objectDict['type'],
                     '-s',
                     '-b'
                     ]

        try:
            messages = subprocess.check_output(pmrep_cmd, stderr=subprocess.STDOUT)
            errors = re.findall('^.*<Warning>.*$|^.*<Error>.*$', messages, re.MULTILINE)
            if len(errors) > 0:
                print(errors)
                # noinspection PyTypeChecker
                return '\n'.join(errors)
        except subprocess.CalledProcessError as e:
            messages = e.output
            print("Error code " + str(e.returncode))
            print("From " + ' '.join(e.cmd))
            print(messages)
            return messages

    # Export only mappings with reusable dependents included
    def exportObjectAutoDependents(self, objectDict, outputPath):
        if objectDict['type'].lower() == 'mapping':
            dependents = True
        else:
            dependents = False
        return self.exportObject(objectDict, dependents, outputPath)

    @staticmethod
    def getFolderName(objectDict):
        return objectDict['type'].capwords() + 's'

    @staticmethod
    def getFileName(objectDict):
        return objectDict['name'] + '.xml'

    @staticmethod
    def attributesString(element):
        s = element.tag
        # print 'attributesString ' + xml.tostring(element)
        if list(element.items()) is not None:
            for attr in sorted(element.items()):
                s += ' ' + ' '.join(attr)
        # print 'end attributesString = ' + s
        return s

    def exportObjectList(self, objectList):
        messageList = list()
        tempDir = tempfile.mkdtemp()
        try:
            newFilesDict = dict()

            self.log.info("{cnt} objects to export".format(cnt=len(objectList)))

            for objectDict in objectList:
                self.log.debug(dict_to_str(objectDict))

                fullTempDir = os.path.join(tempDir, self.getFolderName(objectDict))
                os.makedirs(fullTempDir)
                tempFilePath = os.path.join(fullTempDir, self.getFileName(objectDict))

                self.log.info("Exporting {}/{}".format(self.getFolderName(objectDict),
                                                       self.getFileName(objectDict)
                                                       )
                              )
                try:
                    messages = self.exportObject(objectDict, False, tempFilePath)
                    if messages is not None and len(messages) > 0:
                        messageList.append((self.getFileName(objectDict), messages))

                    targetDir = os.path.join(os.getcwd(), self.getFolderName(objectDict))
                    os.makedirs(targetDir)
                    targetFilePath = os.path.join(targetDir, self.getFileName(objectDict))

                    newFilesDict[targetFilePath] = 1

                    self.log.debug("Copying to {}".format(targetFilePath))
                except NoObjects:
                    pass
        finally:
            # Cleanup temp
            shutil.rmtree(tempDir)
        return messageList

    def validateObjectList(self, objectList):
        messageList = list()

        self.log.info("{cnt} objects to validate".format(cnt=len(objectList)))

        for objectDict in objectList:
            self.log.debug(dict_to_str(objectDict))

            self.log.info("Validating {}/{}".format(self.getFolderName(objectDict),
                                                    self.getFileName(objectDict)
                                                    )
                          )
            try:
                messages = self.validateObject(objectDict)
                if messages is not None and len(messages) > 0:
                    messageList.append((self.getFileName(objectDict), messages))
            except NoObjects:
                pass
        return messageList

    def importXMLFile(self, path, control_file):
        # pmrep objectimport -c "${CONTROL_FILE}" -i "${FILE}" -p
        pmrep_cmd = [
            self.informatica_pmrep(),
            'objectimport',
            '-c', control_file,
            '-i', path,
            '-p',
        ]

        try:
            messages = subprocess.check_output(pmrep_cmd, stderr=subprocess.STDOUT)
            errors = re.findall('^.*<Warning>.*$|^.*<Error>.*$', messages, re.MULTILINE)
            if len(errors) > 0:
                print(errors)
                # noinspection PyTypeChecker
                return '\n'.join(errors)
        except subprocess.CalledProcessError as e:
            messages = e.output
            if e.returncode == 1 and messages.find('No objects to import into repository') != -1:
                messages = "WARNING: No objects to import into repository"
                print(messages)
                return messages
            else:
                print("pmrep Error code " + str(e.returncode))
                print("From " + ' '.join(e.cmd))
                print(messages)
                return messages

    def specifizeControlFile(self, controlFile, workingControlFile):
        with open(controlFile, 'r') as sf:
            with open(workingControlFile, 'w') as tf:
                for line in sf:
                    # Replace generic repository name with our specific one
                    line = re.sub(r'impcntl.dtd',
                                  os.path.join(self.informatica_bin_dir(), 'impcntl.dtd'),
                                  line)

                    tf.write(line)

    def importFileObj(self, fileObj):
        print("Importing {}".format(fileObj.name))
        tempDir = tempfile.mkdtemp()
        scriptDir = os.path.dirname(os.path.realpath(__file__))
        try:
            (_, fileName) = os.path.split(fileObj.name)
            workingFile = os.path.join(tempDir, fileName)
            controlFile = os.path.join(scriptDir, self.control_file_name)
            workingControlFile = os.path.join(tempDir, self.control_file_name)
            self.specifizeControlFile(controlFile, workingControlFile)
            messages = self.importXMLFile(workingFile, workingControlFile)
        except Exception as e:
            messages = e
        finally:
            # Cleanup temp
            shutil.rmtree(tempDir)
        return messages

    def importFile(self, folderName, fileName):
        path = os.path.join(folderName, fileName)
        try:
            with open(path, 'r') as sf:
                messages = self.importFileObj(sf)
        except Exception as e:
            messages = e
        return messages



"""
Created on Sep 17, 2014

@author: Derek Wood
"""

import timeit
from collections import OrderedDict
from datetime import datetime


class Timer(object):

    def __init__(self, task_name: str = None, start_running: bool = True) -> None:
        self.task_name = task_name
        self.stored_time = 0
        self.start_time = None
        self.first_start_time = None
        self.start_time_precise = None
        self.stop_time = None
        self.running = False
        self.format_strings = [
            # (Max_Seconds, Format_String)
            (  60, "{total_seconds:.3f} seconds"),
            (None, "{total_seconds:.1f} seconds ({hours:d}h:{minutes:02d}m:{seconds:02.1f}s)"),
            ]
        if start_running:
            self.start()

    @staticmethod
    def now() -> float:
        return timeit.default_timer()

    @property
    def seconds_elapsed(self) -> float:
        if self.running:
            return self.stored_time + (Timer.now() - self.start_time_precise)
        else:
            return self.stored_time

    @property
    def seconds_elapsed_formatted(self) -> str:
        total_seconds = self.seconds_elapsed
        hours = int(total_seconds / 3600)
        seconds = total_seconds - hours * 3600
        minutes = int(seconds / 60)
        seconds = seconds - minutes * 60
        format_index = 0
        while (format_index + 1) < len(self.format_strings) \
                and self.format_strings[format_index][0] is not None \
                and total_seconds > self.format_strings[format_index][0]:
            format_index += 1
        format_string = self.format_strings[format_index][1]
        return format_string.format(total_seconds=total_seconds,
                                    total_minutes=total_seconds/60,
                                    hours=hours,
                                    minutes=minutes,
                                    seconds=seconds,
                                    )

    @property
    def statistics(self) -> dict:
        stats = OrderedDict()
        if self.first_start_time != self.start_time:
            stats['first start time'] = self.first_start_time
            stats['recent start time'] = self.start_time
        else:
            stats['start time'] = self.start_time
        stats['stop time'] = self.stop_time
        stats['seconds elapsed'] = self.seconds_elapsed_formatted
        if self.task_name is None:
            return stats
        else:
            return dict({self.task_name: stats})

    def message(self, task_name: str = None) -> str:
        if not task_name:
            task_name = self.task_name or "Un-named task"
        if self.running:
            self.stop()
        return "{task} took {secs}".format(task=task_name, secs=self.seconds_elapsed_formatted)

    def message_detailed(self, task_name: str = None) -> str:
        if not task_name:
            task_name = self.task_name or "Un-named task"
        if self.running:
            self.stop()
        return "{task} started at {start} stopped at {stop} and took {secs}"\
            .format(task=task_name,
                    start=self.start_time,
                    stop=self.stop_time,
                    secs=self.seconds_elapsed_formatted
                    )

    def print(self) -> None:
        print(self.message_detailed())

    def start(self) -> None:
        if not self.running:
            self.start_time = datetime.now()
            if self.first_start_time is None:
                self.first_start_time = self.start_time
            self.start_time_precise = Timer.now()
            self.running = True

    def stop(self) -> None:
        if self.running:
            self.stop_time = datetime.now()
            if self.start_time_precise is not None:
                self.stored_time += Timer.now() - self.start_time_precise
            else:
                raise ValueError("stop called on Timer that was not started. Name={}"
                                 .format(self.task_name)
                                 )
            self.running = False

    def reset(self) -> None:
        """Resets the clock statistics and restarts it."""        
        self.start_time = datetime.now()
        self.first_start_time = self.start_time
        self.start_time_precise = Timer.now()
        self.running = True



# https://www.python.org/dev/peps/pep-0563/
from __future__ import annotations

import os.path
import textwrap
import time
from pathlib import Path
from typing import *

import boto3
import sqlalchemy

from bi_etl.bulk_loaders.bulk_loader import BulkLoader
from bi_etl.bulk_loaders.bulk_loader_exception import BulkLoaderException
from bi_etl.bulk_loaders.s3_bulk_load_config import S3_Bulk_Loader_Config
from bi_etl.conversions import strip

if TYPE_CHECKING:
    from bi_etl.scheduler.task import ETLTask
    from bi_etl.components.table import Table


class RedShiftS3Base(BulkLoader):
    def __init__(self,
                 config: S3_Bulk_Loader_Config,
                 ):
        super().__init__(
        )
        self.config = config
        self.s3_user_id = self.config.user_id
        self.s3_region = self.config.region_name
        self.s3_bucket_name = self.config.bucket_name
        self.s3_folder = self.config.folder
        self.s3_files_to_generate = self.config.s3_files_to_generate
        self.s3_file_max_rows = self.config.s3_file_max_rows
        self.s3_clear_before = self.config.s3_clear_before
        self.s3_clear_when_done = self.config.s3_clear_when_done
        self.analyze_compression = self.config.analyze_compression
        self.s3_password = self.config.get_password()
        self.session = boto3.session.Session(
            aws_access_key_id=self.s3_user_id,
            aws_secret_access_key=self.s3_password
        )
        try:
            identity = self.session.client('sts').get_caller_identity()
            self.log.info(f'Account ID = {identity["Account"]}')
            self.log.info(f'Account ARN = {identity["Arn"]}')
            self.log.info(f"Canonical  user = {self.session.client('s3').list_buckets()['Owner']}")
        except Exception as e:
            self.log.warning(e)
        self.s3 = self.session.resource('s3')
        self.bucket = self.s3.Bucket(self.s3_bucket_name)

        self.non_recoverable_error_matches = [
            'The S3 bucket addressed by the query is in a different region from this cluster.',
            'Access denied',
            'The AWS Access Key Id you provided does not exist in our records.'
        ]
        self.lines_scanned_modifier = 0

    def s3_folder_contents(
            self,
            s3_full_folder,
    ):
        return [bucket_object for bucket_object in self.bucket.objects.filter(Prefix=s3_full_folder)]

    def clean_s3_folder(
            self,
            s3_full_folder,
    ):
        self.log.info(f'Cleaning S3 folder {s3_full_folder}')
        for bucket_object in self.bucket.objects.filter(Prefix=s3_full_folder):
            self.log.info(f'Removing {bucket_object}')
            bucket_object.delete()

    def _upload_files_to_s3(
            self,
            local_files: list,
            s3_full_folder: str,
    ):
        if self.s3_clear_before:
            self.clean_s3_folder(s3_full_folder)
        else:
            folder_contents = self.s3_folder_contents(s3_full_folder)
            if folder_contents:
                raise FileExistsError('The target S3 folder is not empty and s3_clear_before = False.'
                                      f'Existing contents = {folder_contents}')

        s3_files = list()
        total_size = 0
        # Upload the files
        for file_number, local_path in enumerate(local_files):
            file_size = os.path.getsize(local_path)
            total_size += file_size
            s3_path = f'{s3_full_folder}/{os.path.basename(local_path)}'
            self.log.info(f"Uploading from '{local_path}' to {self.s3_bucket_name}' key '{s3_path}' size = {file_size:,} bytes")
            s3_files.append(f's3://{self.s3_bucket_name}/{s3_path}')
            self.bucket.upload_file(
                str(local_path),
                str(s3_path),
            )

            response = None
            t_end = time.time() + 60 * 3
            while response is None:
                key = s3_path
                objs = list(self.bucket.objects.filter(Prefix=key))
                if len(objs) > 0 and objs[0].key == key:
                    # self.log.info(f'{key} exists. {t_end}')
                    response = True
                else:
                    if time.time() > t_end:
                        self.log.info(f'{key} is not available for loading. 3 minute wait is over. FATAL. {t_end}')
                        response = False
                    else:
                        self.log.info(f'{key} is probably not finished syncing among AWS S3 nodes, we will retry again after 5 second. {time.time()}')
                        time.sleep(5)
                        response = self.bucket.Object(s3_path).get()

        self.log.info(f"Uploaded {total_size:,} bytes in total to {s3_full_folder}")
        return s3_files

    def _get_error_details(
            self,
            table_object: Table,
            exception: sqlalchemy.exc.SQLAlchemyError
    ) -> BulkLoaderException:
        bulk_loader_exception = BulkLoaderException(exception, password=self.s3_password)

        sql = f"""
                    SELECT *
                    FROM stl_load_errors
                    WHERE query = pg_last_copy_id()
                    ORDER BY starttime DESC
                    """

        self.log.error(f'Checking stl_load_errors with\n{sql}')
        try:
            results = table_object.execute(sql)
        except Exception as e2:
            bulk_loader_exception.add_error(f'Error {e2} when getting stl_load_errors contents')
            results = []
        rows_found = 0
        for row in results:
            rows_found += 1
            filename = strip(row.filename)
            column_name = strip(row.colname)
            c_type = strip(row.type)
            col_length = strip(row.col_length)
            err_reason = str(row.err_reason).strip()
            self.log.error(f'!!!! stl_load_errors row: {rows_found} !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!')
            self.log.error(f'{filename} had an error')
            self.log.error(f'stl_load_errors.err_reason={err_reason}')
            self.log.error(f'stl_load_errors.err_code={row.err_code}')
            self.log.error(f'stl_load_errors.error with column = {column_name} {c_type} {col_length}')
            self.log.error(f'stl_load_errors.raw_field_value = "{str(row.raw_field_value).strip()}"')
            self.log.error(f'stl_load_errors.line number = {row.line_number}')
            self.log.error(f'stl_load_errors.character pos = {row.position}')
            self.log.error(f'stl_load_errors.raw_line={row.raw_line}')
            self.log.error('-' * 80)
            bulk_loader_exception.add_error(f'{column_name} {err_reason}')
        if rows_found == 0:
            bulk_loader_exception.add_error('No records found in stl_load_errors matching file list')
        self.log.error(f'{rows_found} rows found in stl_load_errors for file list')

        return bulk_loader_exception

    def _run_copy_sql(
            self,
            copy_sql: str,
            table_object: Table,
    ) -> int:
        copy_sql = textwrap.dedent(copy_sql)
        sql_safe_pw = copy_sql.replace(self.s3_password, '*' * 8)
        self.log.debug(sql_safe_pw)

        keep_trying = True
        wait_seconds = 15
        t_end = time.time() + wait_seconds

        while keep_trying:
            try:
                table_object.execute(copy_sql)
                # Load worked we can stop looping
                keep_trying = False
            except sqlalchemy.exc.SQLAlchemyError as e:
                safe_e = str(e).replace(self.s3_password, '*' * 8)

                # Check if the error is non-recoverable
                if any(error in str(e) for error in self.non_recoverable_error_matches):
                    self.log.error(f"Cannot recover from this error - {safe_e}")
                    raise

                self.log.error('!' * 80)
                self.log.error(f'!! Details for {safe_e} below:')
                self.log.error('!' * 80)

                bulk_loader_exception = self._get_error_details(
                    table_object=table_object,
                    exception=e,
                )

                if 'S3ServiceException' in str(e):
                    if time.time() > t_end:
                        self.log.info(f'{wait_seconds} seconds wait is over. File cannot be loaded. {t_end}')
                        keep_trying = False
                    self.log.info(f'Will retry in 5 seconds file copy after S3ServiceException failure - {safe_e}')
                    time.sleep(5)
                else:
                    raise bulk_loader_exception

        # TODO: Consider removing this commit the called should do that.
        #       However, requires extensive testing
        table_object.execute(f"COMMIT;")

        sql = f"""
        SELECT SUM(lines_scanned) as lines_scanned, count(DISTINCT filename) as file_cnt
        FROM stl_load_commits
        WHERE query = pg_last_copy_id()
        """
        # Default to -1 rows to indicate error
        rows_loaded = -1
        try:
            results = table_object.execute(sql)
            for row in results:
                lines_scanned = row.lines_scanned
                file_cnt = row.file_cnt
                if self.lines_scanned_modifier == 0:
                    rows_loaded = lines_scanned
                else:
                    rows_loaded = lines_scanned + (self.lines_scanned_modifier * file_cnt)
        except Exception as e:
            e = str(e).replace(self.s3_password, '*' * 8)
            self.log.warning(f"Error getting row count: {e}")

        return rows_loaded

    def get_copy_sql(
            self,
            s3_source_path: str,
            table_to_load: str,
            file_compression: str = '',
            analyze_compression: str = None,
            options: str = '',
    ):
        raise NotImplementedError()

    def load_from_s3_path(
            self,
            s3_source_path: str,
            table_object: Table,
            table_to_load: str = None,
            s3_source_path_is_absolute: bool = True,
            file_list: Optional[List[str]] = None,
            file_compression: str = '',
            options: str = '',
            analyze_compression: str = None,
            perform_rename: bool = False,
    ) -> int:
        if not s3_source_path_is_absolute:
            s3_source_path = f'{self.s3_folder}/{s3_source_path}'

        table_to_load = table_to_load or table_object.qualified_table_name

        copy_sql = self.get_copy_sql(
            s3_source_path=s3_source_path,
            table_to_load=table_to_load,
            file_compression=file_compression,
            analyze_compression=analyze_compression,
            options=options,
        )
        rows_loaded = self._run_copy_sql(
            copy_sql=copy_sql,
            table_object=table_object,
        )
        if perform_rename:
            self.rename_table(table_to_load, table_object)
        return rows_loaded

    def load_from_files(
            self,
            local_files: List[Union[str, Path]],
            table_object: Table,
            table_to_load: str = None,
            perform_rename: bool = False,
            file_compression: str = '',
            options: str = '',
            analyze_compression: str = None,
    ) -> int:
        s3_full_folder = self._get_table_specific_folder_name(self.s3_folder, table_object)
        file_list = self._upload_files_to_s3(local_files, s3_full_folder)

        rows_loaded = self.load_from_s3_path(
            s3_source_path=s3_full_folder,
            file_list=file_list,
            table_object=table_object,
            table_to_load=table_to_load,
            perform_rename=perform_rename,
            file_compression=file_compression,
            options=options,
            analyze_compression=analyze_compression,
        )
        if self.s3_clear_when_done:
            self.clean_s3_folder(s3_full_folder)

        return rows_loaded

    def load_from_iterator(
        self,
        iterator: Iterator,
        table_object: Table,
        table_to_load: str = None,
        perform_rename: bool = False,
        progress_frequency: int = 10,
        analyze_compression: bool = None,
        parent_task: Optional[ETLTask] = None,
    ) -> int:

        raise NotImplementedError()

    def load_table_from_cache(
            self,
            table_object: Table,
            table_to_load: str = None,
            perform_rename: bool = False,
            progress_frequency: int = 10,
            analyze_compression: str = None,
    ) -> int:
        row_count = self.load_from_iterator(
            iterator=table_object.cache_iterator(),
            table_object=table_object,
            analyze_compression=analyze_compression,
        )
        if perform_rename:
            old_name = table_object.qualified_table_name + '_old'
            table_object.execute(f'DROP TABLE IF EXISTS {old_name} CASCADE')
            table_object.execute(f'ALTER TABLE {table_object.qualified_table_name} RENAME TO {old_name}')
            table_object.execute(f'ALTER TABLE {table_to_load} RENAME TO {table_object.qualified_table_name}')
        return row_count



from bi_etl.notifiers.email import Email
from bi_etl.notifiers.jira import Jira
from bi_etl.notifiers.log_notifier import LogNotifier
from bi_etl.notifiers.slack import Slack

NOTIFIER_CLASES = {
        'Email': Email,
        'LogNotifier': LogNotifier,
        'Slack': Slack,
        'Jira': Jira,
    }



# -*- coding: utf-8 -*-
"""
Created on Jan 20, 2016

@author: Derek Wood
"""
from enum import IntEnum, unique


@unique
class RowStatus(IntEnum):
    """
    Row status values
    """
    existing = 1
    insert = 2
    update_whole = 3
    update_partial = 4
    deleted = 5
    unknown = 99



# https://www.python.org/dev/peps/pep-0563/
from __future__ import annotations

import os.path
import tempfile
from pathlib import Path
from typing import *

from bi_etl.bulk_loaders.bulk_loader import BulkLoader
from bi_etl.bulk_loaders.postgresql_bulk_load_config import PostgreSQLBulkLoaderConfig
from bi_etl.components.csv_writer import CSVWriter, QUOTE_MINIMAL
from bi_etl.utility.postgresql.psycopg2_helpers import psycopg2_import_using_cursor

if TYPE_CHECKING:
    from bi_etl.scheduler.task import ETLTask
    from bi_etl.components.table import Table


class PostgreSQLCopy(BulkLoader):
    def __init__(self,
                 config: PostgreSQLBulkLoaderConfig,
                 ):
        super().__init__(
        )
        self.config = config

    def load_from_files(
            self,
            local_files: List[Union[str, Path]],
            table_object: Table,
            table_to_load: str = None,
            perform_rename: bool = False,
            file_compression: str = '',
            options: str = '',
            analyze_compression: str = None,
    ) -> int:
        rows_inserted = 0
        conn = None
        try:
            conn = table_object.table.bind.raw_connection()
            conn.set_client_encoding(self.config.encoding)
            cursor = conn.cursor()
            for file_name in local_files:
                rows_inserted = psycopg2_import_using_cursor(
                    cursor=cursor,
                    table_spec=table_object.qualified_table_name,
                    input_file_path=file_name,
                    delimiter=self.config.delimiter,
                    csv_mode=True,
                    header=self.config.header,
                    null=self.config.null,
                    encoding=self.config.encoding,
                )
                self.log.info(f"{rows_inserted} rows inserted from {file_name}")
            # Note: We need to commit here otherwise the changes are lost when we close the raw connection.
            conn.commit()
        finally:
            if conn is not None:
                conn.close()
        if perform_rename:
            self.rename_table(table_to_load, table_object)
        return rows_inserted

    def load_from_iterator(
            self,
            iterator: Iterator,
            table_object: Table,
            table_to_load: str = None,
            perform_rename: bool = False,
            progress_frequency: int = 10,
            analyze_compression: str = None,
            parent_task: Optional[ETLTask] = None,
    ) -> int:
        with tempfile.TemporaryDirectory(dir=self.config.temp_file_path, ignore_cleanup_errors=True) as temp_dir:
            data_file_path = os.path.join(temp_dir, f'data_{table_object.table_name}.data')

            # Save to a file first, so we can call load_from_files
            with open(data_file_path, 'w+', encoding="utf-8") as file:
                with CSVWriter(
                    parent_task,
                    file,
                    delimiter=self.config.delimiter,
                    column_names=table_object.column_names,
                    include_header=self.config.header,
                    encoding='utf-8',
                    escapechar='\\',
                    quoting=QUOTE_MINIMAL,
                    null=self.config.null,
                ) as writer:
                    for row in iterator:
                        writer.insert_row(row)

            row_count = self.load_from_files(
                [data_file_path],
                table_object=table_object,
                table_to_load=table_to_load,
                perform_rename=perform_rename,
                analyze_compression=analyze_compression,
            )
            return row_count



"""
Created on Sep 15, 2014

@author: Derek Wood
"""
import errno
import importlib
import inspect
import logging
import socket
import traceback
import warnings
from pathlib import Path
from queue import Empty
from typing import *

from config_wrangler.config_templates.sqlalchemy_database import SQLAlchemyDatabase, SQLAlchemyMetadata
from config_wrangler.config_types.dynamically_referenced import DynamicallyReferenced
from pydicti import dicti, Dicti

import bi_etl
import bi_etl.config.notifiers_config as notifiers_config
from bi_etl import utility
from bi_etl.components.etlcomponent import ETLComponent
from bi_etl.config.bi_etl_config_base import BI_ETL_Config_Base, BI_ETL_Config_Base_From_Ini_Env
from bi_etl.database.database_metadata import DatabaseMetadata
from bi_etl.notifiers import LogNotifier, Email, Slack, Jira
from bi_etl.notifiers.notifier_base import NotifierBase
from bi_etl.scheduler import models
from bi_etl.scheduler import queue_io
from bi_etl.scheduler.exceptions import ParameterError, TaskStopRequested
from bi_etl.scheduler.messages import ChildRunOK
from bi_etl.scheduler.messages import ChildRunRequested
from bi_etl.scheduler.messages import ChildSetDisplayName
from bi_etl.scheduler.messages import ChildStatusUpdate
from bi_etl.scheduler.status import Status
from bi_etl.statistics import Statistics
from bi_etl.timer import Timer

if TYPE_CHECKING:
    from bi_etl.scheduler.scheduler_interface import SchedulerInterface

# pylint: disable=too-many-instance-attributes, too-many-public-methods
# pylint: disable=too-many-statements, too-many-branches, too-many-arguments


class ETLTask(object):
    """
    ETL Task runnable base class.

    load() **must** be overridden.

    depends_on() should be overridden.

    start_following_tasks() can be overridden.
    """
    CLASS_VERSION = 1.0
    _task_repo: Dict[str, 'ETLTask'] = dict()

    def __init__(self,
                 config: BI_ETL_Config_Base,
                 task_id=None,
                 parent_task_id=None,
                 root_task_id=None,
                 scheduler=None,
                 task_rec=None,
                 ):
        """
        Constructor. This code will run on the scheduler thread and the execution thread.
        It should do as little as possible.

        Parameters
        ----------
        task_id: int
            The task_id of the job (only for :class:`bi_etl.scheduler.scheduler.Scheduler`).
        parent_task_id: int
            The task_id of the parent of this job (only for :class:`bi_etl.scheduler.scheduler.Scheduler`).
        root_task_id: int
            The task_id of the root ancestor of this job
            (only for :class:`bi_etl.scheduler.scheduler.Scheduler`).
        scheduler: bi_etl.scheduler.scheduler.Scheduler
            The :class:`bi_etl.scheduler.scheduler.Scheduler` this job should be run under.
            Defaults to not running via a scheduler.
        config: bi_etl.config.bi_etl_config_base.BI_ETL_Config_Base
            The configuration :class:`bi_etl.config.bi_etl_config_base.BI_ETL_Config_Base` to use
            (See :doc:`config_ini`).
        """
        self.config = config
        self._log = None
        self.log_file_name = None
        self._task_rec = None

        # If we got both task_id and task_rec
        if task_id and task_rec:
            # Make sure they match
            assert task_id == task_rec.task_id, f"Conflicting task_id values given {task_id} and {task_rec.task_id}"
            self.task_rec = task_rec
        # Otherwise if we got only task_id
        elif task_id:
            assert isinstance(task_id, int), "task_id is not an integer or None!"
            if scheduler:
                # If we got a scheduler, use it to get the task_record
                self.task_rec = scheduler.get_task_record(task_id)
            else:
                # Otherwise make one (it will never get stored though
                self.task_rec = models.ETL_Tasks()
                self.task_id = task_id
        # Otherwise if we got only task_rec
        elif task_rec:
            self.task_rec = task_rec
        # Otherwise we didn't get either (local run)
        else:
            # Make a task_rec (it will never get stored though)
            self.task_rec = models.ETL_Tasks()
        # Make sure we don't refer to task_rec anymore only self.task_rec
        del task_rec

        if self.parent_task_id is None:
            self.parent_task_id = parent_task_id
        else:
            assert parent_task_id == self.task_rec.parent_task_id, f"Conflicting parent_task_id values given {task_id} and {self.task_rec.parent_task_id}"

        if self.root_task_id is None:
            self.root_task_id = root_task_id
        else:
            assert root_task_id == self.task_rec.root_task_id, f"Conflicting parent_task_id values given {root_task_id} and {self.task_rec.root_task_id}"

        self._externally_provided_scheduler = (scheduler is not None)
        self._scheduler = scheduler
        if self.status is None:
            self.status = Status.new
        self._parameters_loaded = False
        self._parameter_dict = dicti()
        self.parent_to_child = None
        self.child_to_parent = None
        self.object_registry = list()
        self.thread_running = None
        self.summary_message_from_client = False
        self.last_log_msg_time = None
        self.pending_log_msgs = list()
        self.warning_messages = set()
        self.last_log_msg = ""
        self.exception = None
        # set initial default value for waiting_for_workflow
        self.waiting_for_workflow = False
        # Try and get waiting_for_workflow from the parent
        if self.parent_task_id is not None:
            try:
                self.waiting_for_workflow = self.parent_task.needs_to_ok_child_runs()
                self.parent_task.register_child(self)
            except TypeError:
                pass  # We're not running on a full Scheduler
        self._children = set()
        self._manager = None
        # Used by the scheduler to tell if this task has written it's dependencies to the log
        self._logged_dependencies = False
        self._normalized_dependents_set = None
        self._mutually_exclusive_with_set = None
        self._database_pool = list()
        self.init_timer = Timer(start_running=False)
        self.load_timer = Timer(start_running=False)
        self.finish_timer = Timer(start_running=False)
        self.suppress_notifications = False
        self.log_handler = None

    def __getstate__(self):
        odict = dict()
        odict['version'] = self.CLASS_VERSION
        odict['task_id'] = self.task_id
        odict['root_task_id'] = self.root_task_id
        odict['parent_task_id'] = self.parent_task_id
        odict['status'] = self.status
        odict['parent_to_child'] = self.parent_to_child
        odict['child_to_parent'] = self.child_to_parent
        odict['_parameter_dict'] = dict(self._parameter_dict)
        odict['config'] = self.config
        # We don't pass scheduler or config from the Scheduler to the running instance
        # odict['scheduler'] = self._scheduler
        # print("__getstate__ {}".format(utility.dict_to_str(odict)))
        return odict

    def __setstate__(self, odict):
        # print("__setstate__ {}".format(utility.dict_to_str(odict)))
        if 'version' not in odict:
            odict['version'] = 0.0
        if odict['version'] != self.CLASS_VERSION:
            raise ValueError("ETLTask versions incompatible between scheduler and target server")
        self.__init__(task_id=odict['task_id'],
                      parent_task_id=odict['parent_task_id'],
                      root_task_id=odict['root_task_id'],
                      config=odict['config'],
                      # We don't pass scheduler from the Scheduler to the running instance
                      # scheduler= odict['scheduler']
                      )
        self.parent_to_child = odict['parent_to_child']
        self.child_to_parent = odict['child_to_parent']
        self._parameter_dict = Dicti(odict['_parameter_dict'])

    def shutdown(self):
        if self._manager is not None:
            self._manager.shutdown()

    def log_logging_level(self):
        # Calling bi_etl.utility version
        utility.log_logging_level(self.log)

    def __repr__(self):
        return f"{self.name}(task_id={self.task_id}, " \
               f"parent_task_id={self.parent_task_id}, " \
               f"root_task_id={self.root_task_id})"

    def __str__(self):
        return self.name

    @property
    def name(self):
        """
        Note: Return value needs to be compatible with find_etl_class
        """
        module = self.__class__.__module__
        return f"{module}.{self.__class__.__name__}"

    @property
    def environment_name(self):
        environment = self.config.bi_etl.environment_name
        if environment == '*qualified_host_name*':
            if self._scheduler is not None:
                environment = self._scheduler.qualified_host_name
            else:
                environment = socket.gethostname()
        return environment

    @property
    def task_rec(self):
        return self._task_rec

    @task_rec.setter
    def task_rec(self, new_value):
        assert isinstance(new_value, models.ETL_Tasks)
        self._task_rec = new_value

    @property
    def status(self):
        return self.task_rec.Status

    @status.setter
    def status(self, new_value):
        self.task_rec.Status = new_value

    @property
    def task_id(self):
        return self.task_rec.task_id

    @task_id.setter
    def task_id(self, new_value):
        if new_value is not None:
            new_value = int(new_value)
        self.task_rec.task_id = new_value

    @property
    def display_name(self):
        return self.task_rec.display_name

    @display_name.setter
    def display_name(self, new_value):
        if self.child_to_parent:
            self.child_to_parent.put(ChildSetDisplayName(self.task_id, new_value))
            self.task_rec.display_name = new_value
        else:
            self.log.debug(f'Setting display_name in task_rec= {new_value}')
            self.task_rec.display_name = new_value
            self.scheduler.session.commit()

    @property
    def parent_task_id(self):
        return self.task_rec.parent_task_id

    @parent_task_id.setter
    def parent_task_id(self, new_value):
        if new_value is not None:
            new_value = int(new_value)
        self.task_rec.parent_task_id = new_value

    @property
    def parent_task(self):
        if self.parent_task_id is not None:
            if self._scheduler is not None and hasattr(self._scheduler, 'get_task_by_id'):
                return self._scheduler.get_task_by_id(self.parent_task_id)  # pylint: disable=no-member
            else:
                raise TypeError('Scheduler required to get ETLTask.parent_task')
        else:
            return None

    @property
    def root_task_id(self):
        return self.task_rec.root_task_id

    @root_task_id.setter
    def root_task_id(self, new_value):
        if new_value is not None:
            new_value = int(new_value)
        self.task_rec.root_task_id = new_value

    @property
    def root_task(self):
        if self.root_task_id is not None:
            from bi_etl.scheduler.scheduler import Scheduler
            if isinstance(self.scheduler, Scheduler):
                assert isinstance(self.scheduler, Scheduler)
                return self.scheduler.get_task_by_id(self.root_task_id)
            else:
                raise TypeError('Scheduler required to get ETLTask.root_task')
        else:
            return None

    @property
    def children(self):
        return self._children

    def register_child(self, child_task_object):
        return self._children.add(child_task_object)

    @property
    def log(self):
        """
        Get a logger using the task name.
        """
        if self._log is None:
            self._log = logging.getLogger(self.name)

        return self._log

    def add_warning(self, warning_message):
        self.warning_messages.add(warning_message)

    # pylint: disable=no-self-use
    def depends_on(self) -> Iterable['ETLTask']:
        """
        Override to provide a static list of tasks that this task will wait on if they are running.

        Each dependent task entry should consist of either
        1) The module name (str)
        2) A tuple of the module name (str) and class name (str)

        This task will run if the dependent jobs are not active at all in the scheduler.
        This does mean you need to be careful with the order that jobs are added to the scheduler since
        if jobs are supposed to run A->B->C, but you add (and commit) job C first, it will see that B is
        not running and start.  The scheduler would then allow A to run, and block B until both A and C
        are is_finished (since it checks forward and backwards for dependencies.
        """
        return list()

    def internal_tasks(self) -> Iterable['ETLTask']:
        """
        Override to provide a static list of tasks that this task will run internally.
        Can be used by the scheduler to build a complete dependency tree.
        """
        return list()

    def dependency_full_set(self, parents: tuple = None) -> FrozenSet['ETLTask']:
        dependency_set = self.depends_on()
        if dependency_set is None:
            dependency_set = set()
        else:
            # Ensure dependency_set is in fact a set.
            # Even if it is a set, copy the set so that we don't modify the one sent by depends_on.
            # It might not always be a unique list object that is safe to modify
            dependency_set = set(dependency_set)

        # Find external dependencies of internal / sub-tasks
        internal_tasks = self.internal_tasks()

        self_tuple = (self,)
        if parents is None:
            parents = self_tuple
        else:
            parents = parents + self_tuple

        for sub_task in internal_tasks:
            if sub_task is None:
                continue
            if not isinstance(sub_task, ETLTask):
                raise ValueError(f"{self}.internal_tasks returned {sub_task} which is not an ETLTask")
            if sub_task not in parents:
                sub_deps = sub_task.dependency_full_set(parents=parents)
                if sub_deps is not None:
                    # Filter for dependencies outside this task
                    for sub_dep in sub_deps:
                        if sub_dep in parents or sub_dep == self:
                            raise ValueError(f"sub_task {sub_task} has dep {sub_dep} overlap with {self} or {parents}")
                        if sub_dep not in internal_tasks:
                            dependency_set.add(sub_dep)
        return frozenset(dependency_set)

    @property
    def normalized_dependents_set(self) -> Set['ETLTask']:
        """
        Build a set of modules this task depends on.
        See depends_on.
        Each will be "normalized" to be a fully qualified name.
        """
        if self._normalized_dependents_set is None:
            normalized_dependents_set = set()
            self._normalized_dependents_set = normalized_dependents_set
            depends_on = self.dependency_full_set()
            if depends_on is not None:
                for dep_value in depends_on:
                    if not isinstance(dep_value, ETLTask):
                        if isinstance(dep_value, str):
                            dep_value = self.PythonDep(dep_value)
                        else:
                            raise ValueError(f"Dependency {dep_value} is not an ETLTask (preferred) or str value")
                    normalized_dependents_set.add(dep_value)

        return self._normalized_dependents_set

    @property
    def target_database(self) -> DatabaseMetadata:
        raise NotImplementedError()

    # noinspection PyPep8Naming
    def PythonDep(self, etl_class_str: str) -> 'ETLTask':
        module, class_name = etl_class_str.rsplit('.', 1)
        mod_object = importlib.import_module(module)
        try:
            class_object = getattr(mod_object, class_name)
        except AttributeError:
            try:
                # First try the whole etl_class_str as a module name
                class_object = importlib.import_module(etl_class_str)
            except ModuleNotFoundError:
                # Next, try case-insensitive search
                found_matches = list()
                class_name_lower = class_name.lower()
                for found_class_name, found_class in inspect.getmembers(mod_object, inspect.isclass):
                    if found_class_name.lower() == class_name_lower:
                        found_matches.append(found_class)
                if len(found_matches) == 0:
                    raise ValueError(f"Module {mod_object} does not contain {class_name}")
                elif len(found_matches) > 1:
                    raise ValueError(
                        f"Module {mod_object} does contains more than one case-insenstive match for {class_name} "
                        f"they are {found_matches}"
                    )
                class_object = found_matches[0]

        if inspect.isclass(class_object):
            if not issubclass(class_object, ETLTask):
                raise ValueError(f"{etl_class_str} resolves to a class of {class_object} which is not a subclass of ETLTask")
        else:
            # class_object is most likely a module.  We'll search it for our class.
            matches = list()
            for class_in_module_name, class_in_module in inspect.getmembers(class_object, inspect.isclass):
                # Check that the class is defined in our module (directly or from a sub-module)
                # and is not imported from elsewhere
                if class_in_module.__module__.startswith(class_object.__name__):
                    if issubclass(class_in_module, ETLTask) and class_in_module != ETLTask:
                        matches.append(class_in_module)
            if len(matches) > 1:
                raise ValueError(
                    f"PythonDep was given a module name '{etl_class_str}' and multiple ETLTask classes found inside it. "
                    f"Matches = {[match.name for match in matches]}"
                )
            elif len(matches) == 0:
                raise ValueError(
                    f"PythonDep was given a module name '{etl_class_str}' and no ETLTask classes found inside it. "
                )
            else:
                class_object = matches[0]

        etl_task = class_object(
            config=self.config,
        )
        return etl_task.get_task_singleton()

    def SQLDep(
            self,
            sql_path: str,
            script_path: str = None,
            database: DatabaseMetadata = None
    ) -> 'bi_etl.utility.run_sql_script.RunSQLScript':
        if database is None:
            try:
                database = self.target_database
            except NotImplementedError:
                pass
        inst = self.get_sql_script_runner(
            database_entry=database,
            script_path=script_path,
            script_name=sql_path,
        )
        return inst.get_task_singleton()

    def _mutually_exclusive_execution(self):
        """
        See mutually_exclusive_execution.
        This method has the default functionality so it's easier to call on that logic when
        overriding mutually_exclusive_execution.
        """
        if self.allow_concurrent_runs():
            return list()
        else:
            return [self.name]

    def mutually_exclusive_execution(self):
        """
        Override to provide a list of task names (or partial names that match modules) 
        that this task can not run at the same time as.

        If allow_concurrent_runs is false, defaults to a list with just self.name
        If allow_concurrent_runs is true, defaults to an empty list
        """
        return self._mutually_exclusive_execution()

    @property
    def mutually_exclusive_with_set(self):
        """
        Build a set of modules this task is mutually exclusive with.
        The list is obtained using `mutually_exclusive_execution`.
        Each list member will be "normalized" to be a fully qualified name.
        """
        if self._mutually_exclusive_with_set is None:
            mutually_exclusive_with_set = set()
            self._mutually_exclusive_with_set = mutually_exclusive_with_set
            mutually_exclusive_with_list = self.mutually_exclusive_execution()
            if mutually_exclusive_with_list is not None:
                for mutually_exclusive_with in mutually_exclusive_with_list:
                    mutually_exclusive_with = self.scheduler.find_etl_classes(mutually_exclusive_with)
                    if len(mutually_exclusive_with) > 0:
                        for mutually_exclusive_with_name in mutually_exclusive_with:
                            mutually_exclusive_with_set.add(mutually_exclusive_with_name)
                    else:
                        self.log.warning(
                            f'mutually_exclusive value {mutually_exclusive_with} '
                            'did not match any classes'
                        )

        return self._mutually_exclusive_with_set

    @property
    def scheduler(self) -> 'SchedulerInterface':
        """
        Get the existing :class`bi_etl.scheduler.scheduler.Scheduler` that this task is running under.
        or
        Get an instance of :class`bi_etl.scheduler.scheduler_interface.SchedulerInterface` that can be
        used to interact with the main Scheduler.
        """
        if self._scheduler is None:
            # Import is done here to prevent circular module level imports
            self.log.debug("Building scheduler")
            from bi_etl.scheduler.scheduler_interface import SchedulerInterface
            self._scheduler = SchedulerInterface(
                config=self.config,
            )
        return self._scheduler

    def add_child_task_to_scheduler(
            self,
            etl_task_class_type,
            parameters=None,
            display_name=None,
        ):
        """
        Start a new task on the :class`bi_etl.scheduler.scheduler.Scheduler`
        that will be a child of this task.
        """
        new_task_id = None
        if self.task_id is not None:
            new_task_id = self.scheduler.add_task_by_class(
                etl_task_class_type,
                parent_task_id=self.task_id,
                root_task_id=self.root_task_id,
                parameters=parameters,
                display_name=display_name,
            )
        else:
            self.log.warning(
                'Not running in a scheduler. '
                f'Child task {etl_task_class_type} not actually scheduled.'
            )
        return new_task_id

    def add_child_task_by_partial_name_to_scheduler(self,
                                                    partial_module_name,
                                                    parameters=None,
                                                    display_name=None,
                                                    ):
        """
        Start a new task on the :class`bi_etl.scheduler.scheduler.Scheduler`
        that will be a child of this task.
        """
        new_task_id = None
        if self.task_id is not None:
            new_task_id = self.scheduler.add_task_by_partial_name(
                partial_module_name,
                parent_task_id=self.task_id,
                root_task_id=self.root_task_id,
                parameters=parameters,
                display_name=display_name,
            )
        else:
            self.log.warning(
                f'Not running in a scheduler. '
                f'Child task {partial_module_name} not actually scheduled.'
            )
        return new_task_id

    def start_following_tasks(self):
        """
        Override to add tasks that should follow after this tasks to the scheduler.
        This is called at the end of ETLTask.run
        """
        return

    def load_parameters(self):
        """
        Load parameters for this task from the scheduler.
        """
        # set to loaded no matter what
        self._parameters_loaded = True
        if self.task_id is not None:
            self.scheduler.load_parameters(self)

    def load_parameters_from_dict(self, parameters: dict):
        self._parameters_loaded = True
        for param_name, param_value in parameters.items():
            self.set_parameter(param_name, param_value, local_only=True)

    def set_parameter(
            self,
            param_name: str,
            param_value: object,
            local_only: bool = False,
            commit: bool = True
    ):
        """
        Add a single parameter to this task.

        Parameters
        ----------
        param_name: str
            The name of the parameter to add
        param_value: object
            The value of the parameter
        commit: bool
        local_only: bool
        """
        if not self._parameters_loaded:
            self.load_parameters()
        self._parameter_dict[param_name] = param_value

        if not local_only:
            if self.task_id is not None:
                self.log.info(f'add_parameter to scheduler {param_name} = {param_value}')
                self.scheduler.add_task_paramter(self.task_id, param_name, param_value, commit=commit)
        else:
            print(f"add_parameter local {param_name}={param_value}")

    # Deprecated name
    def add_parameter(
            self,
            param_name: str,
            param_value: object,
            local_only: bool = False,
            commit: bool = True,
    ):
        warnings.warn("add_parameter is deprecated, please use set_parameter instead")
        self.set_parameter(param_name=param_name,
                           param_value=param_value,
                           local_only=local_only,
                           commit=commit)

    def set_parameters(
            self,
            local_only=False,
            commit=True,
            *args,
            **kwargs
    ):
        """
        Add multiple parameters to this task.
        Parameters can be passed in as any combination of:
        * dict instance. Example ``add_parameters( {'param1':'example', 'param2':100} )``
        * list of lists. Example: ``add_parameters( [ ['param1','example'], ['param2',100] ] )``
        * list of tuples. Example: ``add_parameters( [ ('param1','example'), ('param2',100) ] )``
        * keyword arguments. Example: ``add_parameters(foo=1, bar='apple')``
        
        Parameters
        ----------
        local_only: boolean
            Optional. Default= False. Add parameters to local task only. Do not record in the scheduler.
        commit: boolean 
            Optional. Default= True. Commit changes to the task database.
        args: list
            list of lists. or list of tuples. See above.
        kwargs: dict
            keyword arguments send to parameters. See above.
        """
        # Support add_parameters(param1='example', param2=100)
        self._parameter_dict.update(kwargs)
        for param_name, param_value in kwargs.items():
            self.set_parameter(param_name, param_value, local_only=local_only, commit=commit)

        # Also accept a list of dicts, tuples, or lists
        # eg. add_parameters( [ ('param1','example'), ('param2',100) ] )
        #  or add_parameters( [ ['param1','example'], ['param2',100] ] )
        #  or add_parameters( [ ('param1','example'), ['param2',100] ] )
        #  or parms = {'param1':'example', 'param2':100}
        #     add_parameters(parms)
        #
        # which is equivalent to
        #     add_parameters(**parms)
        for arg in args:
            if isinstance(arg, Mapping):
                for param_name, param_value in arg.items():
                    self.set_parameter(param_name, param_value, local_only=local_only, commit=commit)
            elif hasattr(arg, '__getitem__'):
                if len(arg) == 2:
                    self.set_parameter(arg[0], arg[1], local_only=local_only, commit=commit)
                else:
                    raise ValueError(f"add_parameters sequence {arg} had unexpected length {len(arg)}")

    # Deprecated name
    def add_parameters(self, local_only=False, commit=True, *args, **kwargs):
        warnings.warn("add_parameters is deprecated, please use set_parameters instead")
        self.set_parameters(local_only=local_only,
                            commit=commit,
                            *args,
                            **kwargs)

    def parameters(self):
        """
        Returns a generator yielding tuples of parameter (name,value)
        """
        if not self._parameters_loaded:
            self.load_parameters()
        for param_name in self._parameter_dict:
            yield param_name, self._parameter_dict[param_name]

    def parameter_names(self):
        """
        Returns a list of parameter names
        """
        if not self._parameters_loaded:
            self.load_parameters()
        return list(self._parameter_dict.keys())

    def get_parameter(self, param_name, default=None):
        """
        Returns the value of the parameter with the name provided, or default if that is not None.
        
        Parameters
        ----------
        param_name: str
            The parameter to retrieve        
        default: any 
            The default value. *Default* default = None
            
        Raises
        ------
        ParameterError: 
            If named parameter does not exist and no default is provided.
        """
        if not self._parameters_loaded:
            self.load_parameters()

        try:
            return self._parameter_dict[param_name]
        except KeyError as e:
            if default is None:
                raise ParameterError(e) from e
            else:
                return default

    def add_database(self, database_object):
        # _database_pool is used to close connections when the task finishes
        self._database_pool.append(database_object)

    def get_database_name(self):
        """
        Returns the database name (entry in config) to use for calls to get_database where
        no name is provided.

        :return:
        """
        return NotImplementedError()

    def get_database_metadata(self, db_config: SQLAlchemyDatabase) -> DatabaseMetadata:
        if isinstance(db_config, SQLAlchemyMetadata):
            database_obj = DatabaseMetadata(
                bind=db_config.get_engine(),
                schema=db_config.schema,
            )
        elif isinstance(db_config, SQLAlchemyDatabase):
            database_obj = DatabaseMetadata(
                bind=db_config.get_engine(),
            )
        else:
            raise ValueError(
                "get_database_metadata expects SQLAlchemyDatabase or SQLAlchemyMetadata configs. "
                f"Got {type(db_config)} {db_config}"
            )
        self.add_database(database_obj)
        return database_obj

    def get_database(self, database_name: str) -> DatabaseMetadata:
        db_config = getattr(self.config, database_name)
        return self.get_database_metadata(db_config)

    def get_sql_script_runner(
            self,
            script_name: Union[str, Path],
            script_path: Union[str, Path],
            database_entry: Union[str, DatabaseMetadata, None] = None,
    ) -> 'bi_etl.utility.run_sql_script.RunSQLScript':
        if database_entry is None:
            database_entry = self.get_database_name()
        # Late import to avoid circular dependency
        from bi_etl.utility.run_sql_script import RunSQLScript
        return RunSQLScript(
            config=self.config,
            database_entry=database_entry,
            script_path=script_path,
            script_name=script_name,
        )

    def run_sql_script(
            self,
            script_name: Union[str, Path],
            script_path: Union[str, Path],
            database_entry: Union[str, DatabaseMetadata],
    ):
        runner = self.get_sql_script_runner(
            script_name=script_name,
            script_path=script_path,
            database_entry=database_entry,
        )
        ok = runner.run()
        if not ok:
            raise ValueError(f"{script_name} {runner} failed with error {runner.exception}")

    def register_object(self, obj: Union[ETLComponent, Statistics]):
        """
        Register an ETLComponent or Statistics object with the task.
        This allows the task to 
        1) Get statistics from the component
        2) Close the component when done
        
        """
        self.object_registry.append(obj)
        return obj

    def make_statistics_entry(self, stats_id) -> Statistics:
        stats = Statistics(stats_id=stats_id)
        self.register_object(stats)
        return stats

    # pylint: disable=singleton-comparison
    def debug_sql(self, mode: Union[bool, int] = True):
        """
        Control the output of sqlalchemy engine

        Parameters
        ----------
        mode
            Boolean (debug if True, Error if false) or int logging level.
        """
        if isinstance(mode, bool):
            if mode:
                self.log.info('Setting sqlalchemy.engine to DEBUG')
                logging.getLogger('sqlalchemy.engine').setLevel(logging.DEBUG)
                logging.getLogger('sqlalchemy.engine.base.Engine').setLevel(logging.DEBUG)
            else:
                self.log.info('Setting sqlalchemy.engine to ERROR')
                logging.getLogger('sqlalchemy.engine').setLevel(logging.ERROR)
                logging.getLogger('sqlalchemy.engine.base.Engine').setLevel(logging.ERROR)
        else:
            self.log.info(f'Setting sqlalchemy.engine to {mode}')
            logging.getLogger('sqlalchemy.engine').setLevel(mode)
            logging.getLogger('sqlalchemy.engine.base.Engine').setLevel(mode)

    def __thread_init(self):
        """
        Base class pre-load initialization.  Runs on the execution server.
        Override init instead of this.
        """
        queue_io.redirect_output_to(self.child_to_parent)

        if self.log_file_name is None:
            self.log_file_name = self.name
        self.config.logging.setup_logging()
        self.log_handler = self.config.logging.add_log_file_handler(log_file_prefix=self.log_file_name)

        self.log_logging_level()
        self.log.debug(f"externally_provided_scheduler = {self._externally_provided_scheduler}")

    def init(self):
        """
        pre-load initialization.  Runs on the execution server. Override to add setup tasks.
        
        Note: init method is useful in cases were you wish to define a common base class
        with a single load method. Each inheriting class can then do it's own stuff in init
        With init you can have the flow of execution be:
        
        1) spec_class.init (if any code before super call)
        2) base_class.init
        3) spec_class.init (after super call, where your code should really go)
        4) base_class.load
        
        Note 2: Sometimes the functionality above can be achieved with `__init__`.  However, when
        the scheduler thread creates an ETLTask, it creates an instance and thus runs __init__. 
        Therefore, you will want `__init__` to be as simple as possible.  On the other hand, `init`
        is run only by the task execution thread. So it can safely do more time consuming work. 
        Again though this method is for when class inheritance is used, and that logic can not go 
        into the `load` method.         
        
        Why does the scheduler create an instance?
        It does that in case a task needs a full instance and possibly parameter values in order 
        to answer some of the methods like `depends_on` or `mutually_exclusive_execution`.        
        """
        pass

    def load(self):
        """
        Placeholder for load. This is where the main body of the ETLTask's work should be performed.
        """
        raise AttributeError(f"{self} load not implemented")

    def finish(self):
        """
        Placeholder for post-load cleanup. This might be useful for cleaning up what was done in ``init``.
        It could also allow an inheriting class to begin waiting for children (see ``process_messages``)
        """
        pass

    def send_mesage(self, msg):
        if self.child_to_parent is not None:
            self.child_to_parent.put(msg)

    # noinspection PyMethodMayBeStatic
    def allow_concurrent_runs(self):
        return False

    # noinspection PyMethodMayBeStatic
    def needs_to_ok_child_runs(self):
        """
        Override and return True if you need to give OK before children are allowed to run.
        See process_child_run_requested
        """
        return False

    def process_child_run_requested(self, child_run_requested):
        """
        Override to examine child task before giving OK.
        """
        self.send_mesage(ChildRunOK(child_run_requested.child_task_id))

    # noinspection PyMethodMayBeStatic
    def needs_to_get_child_statuses(self):
        """
        Override and return True if you want to get status updates on children.
        """
        return False

    # noinspection PyMethodMayBeStatic
    def needs_to_get_ancestor_statuses(self):
        """
        Override and return True if you want to get status updates on any ancestor.
        """
        return False

    def process_child_status_update(self, child_status_update):
        """
        Override to examine child task status (ChildRunFinished instances)
        """
        pass

    def process_messages(self, block=False):
        """
        Processes messages for this task.  Should be called somewhere in any row looping.

        **Example Code:**

        .. code-block:: python

            from bi_etl.scheduler.exceptions import WorkflowFinished

            def process_child_status_update(self, childStatusUpdate):
                # Placeholder for real check if done
                example_all_done = self.foo()

                if example_all_done:
                    raise WorkflowFinished()

            def load(self):
                # Placeholder for real load code
                self.load_foo()

                #Begin waiting for children
                try:
                   self.process_messages(block=True)
                except WorkflowFinished:
                    pass
        
        Parameters
        ----------
        block: boolean 
            Block while waiting. Defaults to False.
            If block is True, this will run until a terminating message is received or an exception is thrown by the process_X calls.
            If block if False, you probably want to call inside a loop.
        """
        q = self.parent_to_child
        if q is not None:
            try:
                while True:
                    try:
                        msg = q.get(block=block, timeout=10)
                        self.log.debug(f"process_messages got {msg}")
                        if msg == 'stats':
                            self.child_to_parent.put(self.statistics)
                        elif msg == 'stop':
                            self.log.info("Stop signal received")
                            raise TaskStopRequested()
                        elif isinstance(msg, ChildRunRequested):
                            self.process_child_run_requested(msg)
                        elif isinstance(msg, ChildStatusUpdate):
                            self.process_child_status_update(msg)
                        else:
                            self.log.warning(f"Got unexpected message from parent: {repr(msg)}")
                    except IOError as e:
                        if e.errno == errno.EINTR:
                            continue
                        else:
                            raise
            except Empty:
                pass

    def run(self,
            suppress_notifications=None,
            parent_to_child=None,
            child_to_parent=None,
            handle_exceptions=True,
            ):
        """
        Should not generally be overridden.
        This is called to run the task's code in the init, load, and finish methods.
        """
        self.child_to_parent = child_to_parent
        self.parent_to_child = parent_to_child
        self.__thread_init()

        if suppress_notifications is None:
            # If run directly, assume it a testing run and don't send e-mails
            if self.__class__.__module__ == '__main__':
                self.log.info("Direct module execution detected. Notifications will not be sent")
                self.suppress_notifications = True
        else:
            self.suppress_notifications = suppress_notifications

        self.status = Status.running

        try:
            # Note: init method is useful in cases were you wish to define a common base class
            # with a single load method. Each inheriting class can then do it's own stuff in init
            # With init you can have the flow of execution be:
            #  1) spec_class.init (if any code before super call)
            #  2) base_class.init
            #  3) spec_class.init (after super call, where your code should really go)
            #  3) base_class.load

            self.init_timer.start()
            self.init()
            self.init_timer.stop()

            self.process_messages()

            self.load_timer.start()
            self.load()
            self.load_timer.stop()

            self.process_messages()

            # finish would be the place to cleanup anything done in the init method 
            self.finish_timer.start()
            self.finish()
            self.finish_timer.stop()

            self.log.info(f"{self} done.")
            self.status = Status.succeeded
            stats = self.statistics
            if self.child_to_parent is not None:
                self.child_to_parent.put(stats)
            stats_formatted = Statistics.format_statistics(stats)
            self.log.info(f"{self} statistics=\n{stats_formatted}")

            self.start_following_tasks()
            self.close(error=False)
        except Exception as e:  # pylint: disable=broad-except
            self.close(error=True)
            self.exception = e
            self.status = Status.failed
            if not handle_exceptions:
                raise e
            self.log.exception(e)
            if not self.suppress_notifications:
                environment = self.config.bi_etl.environment_name
                message_list = list()
                message_list.append(repr(e))
                message_list.append(f"Task ID = {self.task_id}")
                if self.config.bi_etl.scheduler is not None:
                    ui_url = self.config.bi_etl.scheduler.base_ui_url
                    if ui_url and self.task_id:
                        message_list.append(f"Run details are here: {ui_url}{self.task_id}")
                message_content = '\n'.join(message_list)
                subject = f"{environment} {self} load failed"

                self.notify(self.config.notifiers.failures, subject=subject, message=message_content,)

            self.log.info(f"{self} FAILED.")
            if self.child_to_parent is not None:
                self.child_to_parent.put(e)
        finally:
            self.config.logging.remove_log_handler(self.log_handler)

        self.log.info(f"Status = {repr(self.status)}")

        # Send out status
        if self.child_to_parent is not None:
            self.child_to_parent.put(self.status)

        return self.status == Status.succeeded

    def get_notifiers(self, channel_list: List[DynamicallyReferenced], auto_include_log=True) -> List[NotifierBase]:
        notifiers_list = list()
        notifier_class_str = 'unset'

        if auto_include_log:
            notifiers_list.append(LogNotifier())

        for config_ref in channel_list:
            config_section = config_ref.get_referenced()
            if not isinstance(config_section, notifiers_config.NotifierConfigBase):
                raise ValueError(f"Notifier is not an instance of NotifierConfigBase: type= {type(config_section)} value= {config_section}")
            try:
                if config_section == 'LogNotifier':
                    notifier_class_str = config_section
                else:
                    notifier_class_str = config_section.notifier_class

                if isinstance(config_section, notifiers_config.LogNotifierConfig):
                    notifier_instance = LogNotifier()
                elif isinstance(config_section, notifiers_config.SMTP_Notifier):
                    notifier_instance = Email(config_section)
                elif isinstance(config_section, notifiers_config.SlackNotifier):
                    notifier_instance = Slack(config_section)
                elif isinstance(config_section, notifiers_config.JiraNotifier):
                    notifier_instance = Jira(config_section)
                else:
                    module, class_name = config_section.notifier_class.rsplit('.', 1)
                    mod_object = importlib.import_module(module)
                    class_object = getattr(mod_object, class_name)
                    notifier_instance = class_object(config_section)

                if notifier_instance is not None:
                    notifiers_list.append(notifier_instance)
            except Exception as e:
                self.log.exception(e)
                if self.config.notifiers.failed_notifications is not None:
                    try:
                        fallback_message = f'Notification to {config_section} {notifier_class_str} failed with error={e}'
                        fallback_notifiers_list = self.get_notifiers(self.config.notifiers.failed_notifications)
                        for fallback_notifier in fallback_notifiers_list:
                            fallback_notifier.send(
                                subject=f"Failed to send to {config_section}",
                                message=fallback_message,
                            )
                            notifiers_list.append(fallback_notifier)
                    except Exception as e:
                        self.log.exception(e)
        return notifiers_list

    def notify(
            self,
            channel_list: List[DynamicallyReferenced],
            subject,
            message=None,
            sensitive_message: str = None,
            attachment=None,
            skip_channels: set = None,
    ):
        if not self.suppress_notifications:
            # Note: all exceptions are caught since we don't want notifications to kill the load
            try:
                filtered_channels = list()

                for channel in channel_list:
                    if skip_channels is None or channel.ref not in skip_channels:
                        filtered_channels.append(channel)

                notifiers_list = self.get_notifiers(filtered_channels)
                for notifier in notifiers_list:
                    try:
                        notifier.send(
                            subject=subject,
                            message=message,
                            sensitive_message=sensitive_message,
                            attachment=attachment,
                        )
                    except Exception as e:
                        self.log.exception(e)
                        if self.config.notifiers.failed_notifications is not None:
                            fallback_message = f"error={e} original_subject={subject} original_message={message}"
                            fallback_notifiers_list = self.get_notifiers(self.config.notifiers.failed_notifications)
                            for fallback_notifier in fallback_notifiers_list:
                                try:
                                    fallback_notifier.send(
                                        subject=f"Failed to send to {notifier}",
                                        message=fallback_message,
                                        sensitive_message=sensitive_message,
                                        attachment=attachment,
                                    )
                                except Exception as e:
                                    self.log.exception(e)

            except Exception as e:
                self.log.exception(e)

    def notify_status(
            self,
            channel_list: List[DynamicallyReferenced],
            status_message: str,
            skip_channels: set = None,
    ):
        """
        Send a temporary status messages that gets overwritten with the next status message that is sent.

        Parameters
        ----------
        channel_list
        status_message
        skip_channels

        Returns
        -------

        """
        if not self.suppress_notifications:
            # Note: all exceptions are caught since we don't want notifications to kill the load
            try:
                filtered_channels = list()

                for channel in channel_list:
                    if skip_channels is None or channel.ref not in skip_channels:
                        filtered_channels.append(channel)

                notifiers_list = self.get_notifiers(filtered_channels)
                for notifier in notifiers_list:
                    try:
                        notifier.post_status(
                            status_message=status_message,
                        )
                    except Exception as e:
                        warnings.warn(repr(e))

            except Exception as e:
                self.log.exception(e)

    @property
    def statistics(self):
        """
        Return the execution statistics from the task and all of it's registered components.
        """
        stats = Statistics(self.name)
        # Only report init stats if something significant was done there
        if self.init_timer.seconds_elapsed > 1:
            stats['Task Init'] = self.init_timer.statistics

        for obj in self.object_registry:
            try:
                name = str(obj)
                # Ensure we capture all distinct object stats by giving each a distinct name
                i = 0
                while name in stats:
                    i += 1
                    name = f"{obj}_{i}"
                stats[name] = obj.statistics
            except AttributeError as e:
                self.log.info(f"'{obj}' does not report statistics. Msg={e}")
            except Exception as e:  # pylint: disable=broad-except
                self.log.exception(e)

        stats['Task Load'] = self.load_timer.statistics

        # Only report finish stats if something significant was done there
        if self.finish_timer.seconds_elapsed > 1:
            stats['Task Finish'] = self.finish_timer.statistics

        return stats

    def close(self, error: bool = False):
        """
        Cleanup the task. Close any registered objects, close any database connections.
        """
        try:
            self.log.debug("close")
            for obj in self.object_registry:
                if hasattr(obj, 'close'):
                    obj.close(error=error)
                del obj
            del self.object_registry
            self.object_registry = list()
            for database in self._database_pool:
                database.bind.dispose()
                database.clear()
            del self._database_pool
            self._database_pool = list()
        except Exception as e:  # pylint: disable=broad-except
            self.log.debug(repr(e))

    def __enter__(self):
        return self

    def __exit__(self, exit_type, exit_value, exit_traceback):
        self.close()

    def get_task_singleton(self):
        inst_name = self.name
        if inst_name in ETLTask._task_repo:
            inst = ETLTask._task_repo[inst_name]
        else:
            ETLTask._task_repo[inst_name] = self
            inst = self
        return inst


##################


def run_task(task_name,
             parameters=None,
             config=None,
             suppress_notifications=None,
             # Scheduler specific
             scheduler=None,
             task_id=None,
             parent_task_id=None,
             root_task_id=None,
             parent_to_child=None,
             child_to_parent=None,
             ):
    """
    Used to find an ETL task module and start it.

    Parameters
    ----------
    task_name: str
        The task name to run. Must match the name or at least *ending* of the name of a module under **etl_jobs**.
    parameters: list or dict
        Parameters for the task. Passed to method :meth:`bi_etl.scheduler.task.ETLTask.add_parameters`.
    config: Union[str, bi_etl.config.bi_etl_config_base.BI_ETL_Config_Base]
        The configuration to use (defaults to reading it from :doc:`config_ini`).
        If passed it should be an string reference to an ini file
        or an instance of :class:`bi_etl.config.bi_etl_config_base.BI_ETL_Config_Base`.
    suppress_notifications: boolean
        Skip sending email on failure? See suppress_notifications in :meth:`bi_etl.scheduler.task.ETLTask.run`.
    scheduler: bi_etl.scheduler.scheduler.Scheduler
        The :class:`bi_etl.scheduler.scheduler.Scheduler` this job should be run under. 
        Optional. Defaults to not running via a scheduler.
    task_id: int
        The task_id of the job (only for :class:`bi_etl.scheduler.scheduler.Scheduler`).
    parent_task_id: int
        The task_id of the parent of this job (only for :class:`bi_etl.scheduler.scheduler.Scheduler`).
    root_task_id: int
        The task_id of the root ancestor of this job (only for :class:`bi_etl.scheduler.scheduler.Scheduler`).
    parent_to_child: Queue
        A queue to use for parent to child communication (only for :class:`bi_etl.scheduler.scheduler.Scheduler`).
    child_to_parent: Queue 
        A queue to use for child to parent communication (only for :class:`bi_etl.scheduler.scheduler.Scheduler`).
    """
    # For memory testing
    # tr = tracker.SummaryTracker()
    # tr.diff()

    try:
        logging.basicConfig(level=logging.DEBUG)
        queue_out_stream = queue_io.redirect_output_to(child_to_parent)
        print("run_task...")
        if config is None:
            config = BI_ETL_Config_Base_From_Ini_Env()
        elif isinstance(config, str):
            config = BI_ETL_Config_Base_From_Ini_Env(file_name=config)
        else:
            print("Using passed config")

        print(f'Scanning for task matching {task_name}')
        # Import is done here to prevent circular module dependency
        from bi_etl.scheduler.scheduler_interface import SchedulerInterface
        if scheduler is None:
            scheduler = SchedulerInterface(config)
        etl_class = scheduler.find_etl_class_instance(task_name)
        etl_task = etl_class(task_id=task_id,
                             parent_task_id=parent_task_id,
                             root_task_id=root_task_id,
                             scheduler=scheduler,
                             config=config,
                             )
        if parameters is not None and len(parameters) > 0:
            etl_task.set_parameters(parameters)
        ran_ok = etl_task.run(suppress_notifications=suppress_notifications,
                              parent_to_child=parent_to_child,
                              child_to_parent=child_to_parent,
                              )
        # print(f"ran_ok = {ran_ok}")
        etl_task.close()
        print("run_task is done")

        # For memory testing
        # gc.collect()
        # tr.print_diff()

    except Exception as e:
        traceback.print_exc()
        print((repr(e)))
        if child_to_parent is not None:
            child_to_parent.put(e)
        raise e

    queue_io.restore_standard_output()
    return ran_ok



# https://www.python.org/dev/peps/pep-0563/
from __future__ import annotations

import os.path
import os.path
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from tempfile import TemporaryDirectory
from typing import *

import fastavro

from bi_etl.bulk_loaders.redshift_s3_base import RedShiftS3Base
from bi_etl.bulk_loaders.s3_bulk_load_config import S3_Bulk_Loader_Config

if TYPE_CHECKING:
    from bi_etl.components.table import Table
    from bi_etl.scheduler.task import ETLTask


class RedShiftS3AvroBulk(RedShiftS3Base):
    def __init__(
        self,
        config: S3_Bulk_Loader_Config,
    ):
        super().__init__(
            config=config,
        )
        if self.s3_file_max_rows is None and self.s3_files_to_generate is None:
            self.s3_file_max_rows = 50000

        # Redshift appears to count the schema as a line read
        self.lines_scanned_modifier = -1

        try:
            import snappy
            self.codec = 'snappy'
        except ImportError:
            self.codec = 'deflate'

    @property
    def needs_all_columns(self):
        return False

    def get_copy_sql(
            self,
            s3_source_path: str,
            table_to_load: str,
            file_compression: str = '',
            analyze_compression: str = None,
            options: str = '',
    ):
        analyze_compression = analyze_compression or self.analyze_compression
        if analyze_compression:
            options += f' COMPUPDATE {self.analyze_compression} '

        return f"""\
                COPY {table_to_load} FROM 's3://{self.s3_bucket_name}/{s3_source_path}'                      
                     credentials 'aws_access_key_id={self.s3_user_id};aws_secret_access_key={self.s3_password}'
                     AVRO 'auto'
                     {file_compression}  
                     {options}
               """

    def _write_to_file(
            self,
            temp_dir,
            file_number: int,
            parsed_schema,
            data_chunk: list,
            local_files_list: list,
    ):

        self.log.debug(f"Starting bulk load chunk {file_number}")
        filepath = os.path.join(temp_dir, f'data_{file_number}.avro')
        local_files_list.append(filepath)
        with open(filepath, 'wb') as avro_file:
            # avro_iterators = self.distribute(iterator, writer_pool_size)
            fastavro.writer(avro_file, parsed_schema, data_chunk, codec=self.codec)
        self.log.debug(f"Closed bulk load chunk {file_number}: {filepath}")

    def load_from_iterator(
           self,
           iterator: Iterator,
           table_object: Table,
           table_to_load: str = None,
           perform_rename: bool = False,
           progress_frequency: int = 10,
           analyze_compression: str = None,
           parent_task: Optional[ETLTask] = None,
    ) -> int:

        table_to_load = table_to_load or table_object.qualified_table_name

        with TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
            # Generate schema from table
            fields = list()
            type_map = {
                str: ['string'],
                bytes: ['bytes'],
                date: ['string'],
                datetime: ['string'],
                time: ['string'],
                timedelta: ['string'],
                bool: ['boolean'],
                int: ['long'],
                float: ['double'],
                Decimal: ['int', 'double'],
            }
            for column in table_object.columns:
                if column.type.python_type in type_map:
                    col_type = type_map[column.type.python_type]
                else:
                    col_type = ['string']

                fields.append({
                    'name': column.name,
                    'type': ["null"] + col_type,
                    'default': None,
                })

            schema = {
                "name": table_to_load,
                "type": "record",
                'fields': fields,
            }

            self.log.debug(f"schema = {schema}")

            parsed_schema = fastavro.parse_schema(schema)

            self.log.debug(f"parsed_schema = {parsed_schema}")

            file_number = 0
            data_length = 0
            local_files_list = list()

            # Note: Tried using this method to chunk the list, but it didn't work with the
            # gevent based threading used by insert_row with bulk loader
            # for data_chunk in more_itertools.chunked(iterator, self.s3_file_max_rows):

            data_chunk = list()
            for row in iterator:
                data_chunk.append(row)

                if len(data_chunk) >= self.s3_file_max_rows:
                    data_length += len(data_chunk)
                    file_number += 1
                    self._write_to_file(
                        temp_dir=temp_dir,
                        file_number=file_number,
                        parsed_schema=parsed_schema,
                        data_chunk=data_chunk,
                        local_files_list=local_files_list,
                    )
                    data_chunk = list()

            # Write final chunk
            if len(data_chunk) >= 0:
                data_length += len(data_chunk)
                file_number += 1
                self._write_to_file(
                    temp_dir=temp_dir,
                    file_number=file_number,
                    parsed_schema=parsed_schema,
                    data_chunk=data_chunk,
                    local_files_list=local_files_list,
                )

            self.log.debug(f"Loading from files {local_files_list}")
            rows_loaded = self.load_from_files(
                local_files_list,
                table_object=table_object,
                table_to_load=table_to_load,
                perform_rename=perform_rename,
                analyze_compression=analyze_compression,
            )
            if rows_loaded != data_length:
                self.log.error(f"COPY from files should have loaded {data_length:,} but it reports {rows_loaded:,} rows loaded")
            else:
                self.log.info(f"{self} had nothing to do with 0 rows found")
            return data_length



from pathlib import Path
from typing import Optional

from config_wrangler.config_templates.config_hierarchy import ConfigHierarchy


class PostgreSQLBulkLoaderConfig(ConfigHierarchy):
    temp_file_path: Optional[Path] = None
    delimiter: str = '|'
    header: bool = True
    null: str = ''
    encoding: str = 'utf-8'



"""
Created on Feb 27, 2015

@author: Derek Wood
"""
# https://www.python.org/dev/peps/pep-0563/
from __future__ import annotations

import typing
from datetime import datetime

from sqlalchemy.sql import Selectable
from sqlalchemy.sql.expression import bindparam

from bi_etl.components.row.row import Row
from bi_etl.config.bi_etl_config_base import BI_ETL_Config_Base
from bi_etl.conversions import ensure_datetime
from bi_etl.exceptions import BeforeAllExisting, AfterExisting, NoResultFound
from bi_etl.lookups.lookup import Lookup
from bi_etl.utility import dict_to_str

if typing.TYPE_CHECKING:
    from bi_etl.components.etlcomponent import ETLComponent


__all__ = ['RangeLookup']


class RangeLookup(Lookup):
    def __init__(self,
                 lookup_name: str,
                 lookup_keys: list,
                 parent_component: ETLComponent,
                 begin_date,
                 end_date,
                 config: BI_ETL_Config_Base = None,
                 use_value_cache: bool = True,
                 ):
        super().__init__(
            lookup_name=lookup_name,
            lookup_keys=lookup_keys,
            parent_component=parent_component,
            use_value_cache=use_value_cache,
            config=config
        )
        self.begin_date = begin_date
        self.end_date = end_date
        self._remote_lookup_stmt_no_date = None
        self._len = 0

    def __len__(self):
        return self._len

    def cache_row(
            self,
            row: Row,
            allow_update: bool = True,
            allow_insert: bool = True,
    ):
        """
        Adds the given row to the cache for this lookup.

        Parameters
        ----------
        row: Row
            The row to cache

        allow_update: boolean
            Allow this method to update an existing row in the cache.

        allow_insert: boolean
            Allow this method to insert a new row into the cache

        Raises
        ------
        ValueError
            If allow_update is False and an already existing row (lookup key) is passed in.

        """
        if self.cache_enabled:
            assert isinstance(row, Row), "cache_row requires Row and not {}".format(type(row))

            if self.use_value_cache:
                self._update_with_value_cache(row)

            lk_tuple = self.get_hashable_combined_key(row)
            if self._cache is None:
                self.init_cache()                
            versions_collection = self._cache.get(lk_tuple, None)
            effective_date = ensure_datetime(row[self.begin_date])
            assert isinstance(effective_date, datetime)
                
            if versions_collection is None:
                versions_collection = Lookup.VERSION_COLLECTION_TYPE()

            if effective_date in versions_collection:
                if not allow_update:
                    self.log.error('Key already in lookup!')
                    self.log.error('Existing row = {}'.format(repr(versions_collection[effective_date])))
                    self.log.error('New duplicate row = {}'.format(repr(row)))
                    raise ValueError('Key {} + date {} already in cache and allow_update was False'.format(
                        lk_tuple,
                        effective_date
                    ))
            else:
                self._len += 1
            
            versions_collection[effective_date] = row
            # TODO: Move above into if versions_collection is None so it's done only once
            self._cache[lk_tuple] = versions_collection
            
            # Capture memory usage snapshots
            if self._row_size is None:   
                self._get_first_row_size(row)
            else:
                self.check_estimate_row_size()
            
    def uncache_row(self, row: Lookup.ROW_TYPES):
        if isinstance(row, tuple) or isinstance(row, list):
            raise ValueError("{}.uncache_row requires a Row not a tuple since it needs the date".format(self.__class__.__name__))
        else:
            lk_tuple = self.get_hashable_combined_key(row)
        if self._cache is not None:
            versions_collection = self._cache.get(lk_tuple, None)
            effective_date = ensure_datetime(row[self.begin_date])
                
            if versions_collection:                
                # Look for and remove existing instance that are exactly the same date
                try:
                    del versions_collection[effective_date]
                except (KeyError, ValueError):
                    # Not found, that's OK
                    pass

    def uncache_set(self, row: Lookup.ROW_TYPES):
        if self._cache is not None:
            if isinstance(row, self._hashable_key_type):
                lk_tuple = row
            else:
                lk_tuple = self.get_hashable_combined_key(row)
            del self._cache[lk_tuple]

    def __iter__(self):
        """
        The natural keys will come out in any order. However, the versions within a natural key set will come out in ascending order.  
        """
        if self._cache is not None:
            for versions_collection in list(self._cache.values()):
                for row in versions_collection.values():
                    yield row

    def get_versions_collection(self, row: Lookup.ROW_TYPES) -> typing.MutableMapping[datetime, Row]:
        """
        This method exists for compatibility with range caches

        Parameters
        ----------
        row
            The row with keys to search row

        Returns
        -------
        A MutableMapping of rows
        """

        if isinstance(row, self._hashable_key_type):
            lk_tuple = row
        else:
            lk_tuple = self.get_hashable_combined_key(row)

        if not self.cache_enabled:
            raise ValueError("Lookup {} cache not enabled".format(self.lookup_name))
        if self._cache is None:
            self.init_cache()

        try:
            versions_collection = self._cache[lk_tuple]
            return versions_collection
        except KeyError:
            raise NoResultFound()

    def find_in_cache(self, row: Lookup.ROW_TYPES, **kwargs) -> Row:
        """
        Find an existing row in the cache effective on the date provided.
        Can raise ValueError if the cache is not setup.
        Can raise NoResultFound if the key is not in the cache.
        Can raise BeforeAllExisting is the effective date provided is before all existing records.
        """
        effective_date = kwargs.get('effective_date')
        if effective_date is None:
            effective_date = ensure_datetime(row[self.begin_date])

        versions_collection = self.get_versions_collection(row)
        try:
            return self._get_version(versions_collection, effective_date)
        except TypeError as e:
            msg = str(e)
            msg += "\nversions_collection = {}".format(versions_collection)
            # noinspection PyTypeChecker
            msg += "\nversions_collection key[0]= {}".format(next(versions_collection.keys()))
            msg += "\neffective_date = {}".format(effective_date)
            raise TypeError(msg)
                
    def _add_remote_stmt_where_clause(self, stmt: Selectable) -> Selectable:
        stmt = super(RangeLookup, self)._add_remote_stmt_where_clause(stmt)
        # noinspection PyUnresolvedReferences
        stmt = stmt.where(bindparam('eff_date') >= self.parent_component.get_column(self.begin_date))
        # noinspection PyUnresolvedReferences
        stmt = stmt.where(bindparam('eff_date') <= self.parent_component.get_column(self.end_date))
        return stmt

    def _get_remote_stmt_where_values(
            self,
            row: Lookup.ROW_TYPES,
            effective_date: datetime = None,
            ) -> dict:
        values_dict = super(RangeLookup, self)._get_remote_stmt_where_values(row)
        if effective_date is None:
            if isinstance(row, tuple) or isinstance(row, list):
                effective_date_value_name = 'k{}'.format(len(row))
                effective_date = ensure_datetime(values_dict[effective_date_value_name])
                del values_dict[effective_date_value_name]
            else:
                effective_date = ensure_datetime(row[self.begin_date])
        values_dict['eff_date'] = effective_date
        return values_dict
    
    def find_in_remote_table(
            self,
            row: Lookup.ROW_TYPES,
            **kwargs
            ) -> Row:
        """
        Find a matching row in the lookup based on the lookup index (keys)
        
        Only works if parent_component is based on bi_etl.components.readonlytable
        """

        effective_date = kwargs.get('effective_date')
        
        if self._remote_lookup_stmt is None:
            stmt = self.parent_component.select()           
            stmt = self._add_remote_stmt_where_clause(stmt)
            self._remote_lookup_stmt = stmt.compile()
            # Build a second statement that does the lookup without dates
            # Which is what our parent Lookup does so we call that
            stmt_no_date = self.parent_component.select()
            stmt_no_date = super(RangeLookup, self)._add_remote_stmt_where_clause(stmt_no_date)
            # order by the begin date
            stmt_no_date = stmt_no_date.order_by(self.parent_component.get_column(self.begin_date))
            self._remote_lookup_stmt_no_date = stmt_no_date.compile()

        values_dict = self._get_remote_stmt_where_values(row, effective_date)

        rows = list(self.parent_component.execute(self._remote_lookup_stmt, values_dict))
        if len(rows) == 0:
            # Use the second statement that does the lookup without dates
            # Which is what our parent Lookup does so we call that to set values
            values_dict = super(RangeLookup, self)._get_remote_stmt_where_values(row)
            results = self.parent_component.execute(self._remote_lookup_stmt_no_date, values_dict)
            all_key_rows = results.fetchall()
            if len(all_key_rows) == 0:
                raise NoResultFound()
            else:
                row_counter = 0
                for row in all_key_rows:
                    row_counter += 1
                    if row_counter == 1 and effective_date < row[self.begin_date]:
                        raise BeforeAllExisting(self.parent_component.Row(row), effective_date=row[self.begin_date])
                    elif (ensure_datetime(row[self.begin_date])
                          >= ensure_datetime(effective_date)
                          >= ensure_datetime(row[self.end_date])
                          ):
                        return row
                # If we reach this point return the last row in an AfterExisting exception
                raise AfterExisting(prior_row=row, effective_date=effective_date)

        elif len(rows) == 1:
            return self.parent_component.Row(rows[0])
        else:
            msg = 'dict_to_str statement {}\n'.format(self._remote_lookup_stmt)
            msg += 'using keys {}\n'.format(dict_to_str(values_dict))
            msg += 'matched multiple records: \n'
            row_cnt = 0
            for r in rows:
                row_cnt += 1
                if row_cnt >= 10:
                    msg += '...'
                    break
                msg += dict_to_str(self.get_list_of_lookup_column_values(r), depth=2)
                msg += '\n--------------------------\n' 
            raise RuntimeError(msg)

    def find_versions_list_in_remote_table(self, row: Lookup.ROW_TYPES) -> list:
        """
        Find a matching row in the lookup based on the lookup index (keys)

        Only works if parent_component is based on bi_etl.components.readonlytable
        """
        self.stats.timer.start()
        if self._remote_lookup_stmt_no_date is None:
            # noinspection PyUnresolvedReferences
            stmt = self.parent_component.select()
            # Call parent _add_remote_stmt_where_clause so that it doesn't add the date filter
            stmt = super()._add_remote_stmt_where_clause(stmt)
            stmt = stmt.order_by(self.begin_date)
            self._remote_lookup_stmt_no_date = stmt.compile()

        values_dict = super()._get_remote_stmt_where_values(row)

        # noinspection PyUnresolvedReferences
        result = list(self.parent_component.execute(self._remote_lookup_stmt_no_date, values_dict))
        rows = list()
        for result_proxy_row in result:
            row = self.parent_component.Row(data=result_proxy_row)
            rows.append(row)
        self.stats.timer.stop()
        if len(rows) == 0:
            raise NoResultFound()
        else:
            return rows



"""
Created on Aug 28, 2014

@author: Derek Wood
"""
from bi_etl.components.table import Table
from bi_etl.config.bi_etl_config_base import BI_ETL_Config_Base, BI_ETL_Config_Base_From_Ini_Env
from bi_etl.scheduler.task import ETLTask, Status


class ETL_Task_Status_CD(ETLTask):
    def __init__(self,
                 task_id=None,
                 parent_task_id=None,
                 root_task_id=None,
                 scheduler=None,
                 task_rec=None,
                 config: BI_ETL_Config_Base = None,
                 ):
        if config is None:
            config = BI_ETL_Config_Base_From_Ini_Env()
        super().__init__(
            config=config,
            task_id=task_id,
            parent_task_id=parent_task_id,
            root_task_id=root_task_id,
            scheduler=scheduler,
            task_rec=task_rec,
        )
        assert isinstance(self.config, BI_ETL_Config_Base)

    def depends_on(self):
        return [
        ]

    def load(self):
        log = self.log

        log.info("Connecting to scheduler database")
        database = self.get_database_metadata(self.config.bi_etl.scheduler.db)

        with Table(self,
                   database,
                   'etl_task_status_cd',
                   delete_flag='delete_flag',
                   track_source_rows=True,
                   ) as etl_task_status_cd:
            # cdlist.trace = True
            # etl_task_status_cd.trace_data = True
            etl_task_status_cd.fill_cache()

            iteration_header = etl_task_status_cd.full_iteration_header
            for status in Status:
                row = etl_task_status_cd.Row(iteration_header=iteration_header)
                row['status_id'] = status.value
                status_name = status.name.replace('_', ' ').title()
                status_name = status_name.replace('Cpu', 'CPU')
                row['status_name'] = status_name
                # Add delete_flag to source
                row['delete_flag'] = 'N'

                # etl_task_status_cd.trace_data = True
                # self.debug_sql(True)
                etl_task_status_cd.upsert(
                    row,
                )

            etl_task_status_cd.commit()

            # Process deletes
            log.info("Checking for deletes")
            logically_deleted = etl_task_status_cd.Row()
            logically_deleted['delete_flag'] = 'Y'
            etl_task_status_cd.update_not_processed(logically_deleted)

            etl_task_status_cd.commit()
        log.info("Done")


if __name__ == '__main__':
    task = ETL_Task_Status_CD()
    task.run(suppress_notifications=True)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bi_etl',
 'bi_etl.boto3_helper',
 'bi_etl.bulk_loaders',
 'bi_etl.components',
 'bi_etl.components.get_next_key',
 'bi_etl.components.row',
 'bi_etl.config',
 'bi_etl.database',
 'bi_etl.exceptions',
 'bi_etl.informatica',
 'bi_etl.lookups',
 'bi_etl.notifiers',
 'bi_etl.parallel',
 'bi_etl.parameters',
 'bi_etl.performance_test',
 'bi_etl.scheduler',
 'bi_etl.scheduler.scheduler_etl_jobs',
 'bi_etl.test_notebooks',
 'bi_etl.utility',
 'bi_etl.utility.postgresql',
 'bi_etl.utility.sql_server']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.35',
 'btrees>=4.7.2',
 'config-wrangler>=0.5.0',
 'gevent>=21.8.0',
 'more-itertools>=9.0.0',
 'openpyxl>=3.0.5',
 'psutil>=5.7.2',
 'pydicti>=1.1.6',
 'semidbm>=0.5.1',
 'sqlparse>=0.4.2']

extras_require = \
{':extra == "keyring"': ['keyring>=21.1.0'],
 'jira': ['jira>=3.1.1'],
 'scheduler': ['pyramid>=1.10.4', 'pyjtable>=0.1.1'],
 'slack': ['slack-sdk>=3.19.5'],
 'test': ['psycopg2>=2.8.6']}

setup_kwargs = {
    'name': 'bi-etl',
    'version': '1.4',
    'description': 'Python ETL Framework',
    'long_description': '# bi_etl Python ETL Framework for BI\n\n## Docs\n\n[Please see the documentation site for detailed documentation.](https://bietl.dev/bi_etl/)\n\nPython based ETL (Extract Transform Load) framework geared towards BI databases in particular. \nThe goal of the project is to create reusable objects with typical technical transformations used in loading dimension tables.\n\n## Guiding Design Principles\n1. Don’t Repeat Yourself (DRY).\n\n1. The source or target of an ETL owns the metadata (list of columns and data types). The ETL generally has no reason to define those again unless the ETL requires a change. If a datatype must be changed, only that one column’s new type should be specified. If a column name must be changed, only the source & target column names that differ should be specified.\n\n1. Data Quality is King\n\n1. Data quality is more important than performance. For example, the process should fail before truncating data contents (i.e. loading 6 characters into a 5 character field) even if that means sacrificing some load performance.\n\n1. Give helpful error messages.\n\n1. Make it as easy as possible to create re-usable modules.\n\n1. SQL is a very powerful transformation language. The Transform Extract Load (TEL) model should be supported.',
    'author': 'Derek Wood',
    'author_email': 'bietl_info@bietl.dev',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/arcann/bi_etl',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import logging

from sqlalchemy import Column
from sqlalchemy import event
from sqlalchemy import exc
from sqlalchemy.dialects.oracle.base import NCLOB
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import backref
from sqlalchemy.orm import relationship
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.collections import attribute_mapped_collection
from sqlalchemy.pool import Pool
from sqlalchemy.sql.expression import text
from sqlalchemy.sql.schema import ForeignKey
from sqlalchemy.sql.schema import MetaData
from sqlalchemy.sql.schema import Sequence
from sqlalchemy.sql.sqltypes import BLOB
from sqlalchemy.sql.sqltypes import Unicode
from sqlalchemy.types import DateTime
from sqlalchemy.types import INTEGER
from sqlalchemy.types import VARCHAR

from bi_etl.config.bi_etl_config_base import BI_ETL_Config_Base_From_Ini_Env
from bi_etl.scheduler.status import Status
from bi_etl.utility import dict_to_str

log = logging.getLogger(__name__)


# pylint: disable=missing-docstring, too-few-public-methods
# pylint: disable=unused-argument

@event.listens_for(Pool, "checkout")
def ping_connection(dbapi_connection, connection_record, connection_proxy):
    """
    Test connections before they are used (finds disconnected sessions)

    Parameters
    ----------
    dbapi_connection
    connection_record
    connection_proxy

    Returns
    -------

    """
    cursor = dbapi_connection.cursor()
    try:
        # Only ping connections that support it
        if hasattr(dbapi_connection, 'ping'):
            dbapi_connection.ping()
    except:
        # optional - dispose the whole pool
        # instead of invalidating one at a time
        # connection_proxy._pool.dispose()

        # raise DisconnectionError - pool will try
        # connecting again up to three times before raising.
        raise exc.DisconnectionError()
    cursor.close()


# Only use ZopeTransactionExtension on the web tier,
# on the scheduler the import will likely fail, but we don't use DBSession there
try:
    from zope.sqlalchemy import ZopeTransactionExtension  # @UnresolvedImport

    DBSession = scoped_session(sessionmaker(extension=ZopeTransactionExtension()))
except ImportError:
    ZopeTransactionExtension = None
    DBSession = None

metadata = MetaData(schema='public')
Base = declarative_base(metadata=metadata)


class ETL_Scheduler(Base):
    __tablename__ = 'etl_scheduler'
    __table_args__ = {'quote': False}
    scheduler_id = Column(INTEGER, primary_key=True)
    host_name = Column(VARCHAR(400), nullable=False)
    qualified_host_name = Column(VARCHAR(400), nullable=False)
    last_heartbeat = Column(DateTime, nullable=True)


class ETL_Tasks(Base):
    __tablename__ = 'etl_tasks'
    __table_args__ = {'quote': False}
    task_id = Column(INTEGER, Sequence('etl_task_id_seq'), primary_key=True)
    scheduler_id = Column(INTEGER, ForeignKey('etl_scheduler.scheduler_id'), nullable=False)
    scheduler = relationship('ETL_Scheduler')
    submitted_date = Column(DateTime, nullable=False, server_default=text("sysdate"))
    status_id = Column(INTEGER, ForeignKey('etl_task_status_cd.status_id'), nullable=False, )
    status_cd = relationship('ETL_Task_Status_CD', lazy="joined", )
    submit_by_user_id = Column(VARCHAR(100), nullable=True)
    pid = Column(INTEGER, nullable=True)
    modulename = Column(VARCHAR(400), nullable=False)
    classname = Column(VARCHAR(400), nullable=True)
    display_name = Column(VARCHAR(400), nullable=True)
    parent_task_id = Column(INTEGER, ForeignKey('etl_tasks.task_id'), nullable=True)
    root_task_id = Column(INTEGER, ForeignKey('etl_tasks.task_id'), nullable=True)
    children = relationship("ETL_Tasks",
                            foreign_keys=[parent_task_id],
                            backref=backref('parent_task', remote_side=[task_id])
                            )
    ancestors = relationship("ETL_Tasks",
                             foreign_keys=[root_task_id],
                             backref=backref('root_task', remote_side=[task_id])

                             )
    started_date = Column(DateTime, nullable=True)
    finished_date = Column(DateTime, nullable=True)
    summary_message = Column(Unicode(2000), nullable=True)
    parameters = relationship('ETL_Task_Params')
    log_entries = relationship('ETL_Task_Log', order_by="ETL_Task_Log.log_entry_ts")
    stats = relationship('ETL_Task_Stats',
                         collection_class=attribute_mapped_collection('stat_name'),
                         cascade="all, delete-orphan"
                         )

    def __repr__(self):
        return "ETL_Tasks({})".format(dict_to_str(self))

    def __str__(self):
        return "ETL_Tasks(task_id={})".format(self.task_id)

    @property
    def Status(self):
        if self.status_id is not None:
            return Status(self.status_id)
        else:
            return None

    @Status.setter
    def Status(self, status_object):
        try:
            self.status_id = status_object.value
        except AttributeError:
            self.status_id = int(status_object)

    @property
    def children_id_list(self):
        c_lst = [str(c.task_id) for c in self.children]
        return ','.join(c_lst)

    @staticmethod
    def get_next_task_id(session):
        # We have to query the sequence each time
        # because this code might be running on a different server or thread from others
        next_id = list(session.execute(ETL_Tasks.task_id.default))[0][0]
        return next_id

    def errors(self):
        result = list()
        for log_row in self.log_entries:
            if log_row.log_entry is not None:
                lines = str(log_row.log_entry).split('\n')
                for line in lines:
                    if line.startswith('ERROR'):
                        result.append(line)
        return result


class ETL_Task_Dependency(Base):
    __tablename__ = 'etl_task_dependency'
    __table_args__ = {'quote': False}
    task_id = Column(INTEGER, ForeignKey('etl_tasks.task_id'), primary_key=True, )
    dependent_on_task_id = Column(INTEGER, ForeignKey('etl_tasks.task_id'), primary_key=True, )
    dependent_reason = Column(VARCHAR(100), nullable=False)
    current_blocking_flag = Column(VARCHAR(1), nullable=False)

    task = relationship(ETL_Tasks,
                        foreign_keys=[task_id],
                        backref=backref("dependencies", )
                        )
    dependent_on_task = relationship(ETL_Tasks,
                                     foreign_keys=[dependent_on_task_id],
                                     backref=backref("depends_on_this", )
                                     )

    def __init__(self, task, dependent_on_task, dependent_reason):
        self.task = task
        self.dependent_on_task = dependent_on_task
        self.dependent_reason = str(dependent_reason)
        self.current_blocking_flag = 'Y'
        super().__init__()

    def __str__(self):
        return "(id={},reason={},current_blocking_flag={})".format(self.dependent_on_task, self.dependent_reason,
                                                                   self.current_blocking_flag)


class ETL_Task_Status_CD(Base):
    __tablename__ = 'etl_task_status_cd'
    __table_args__ = {'quote': False}
    status_id = Column(INTEGER, primary_key=True, )
    status_name = Column(VARCHAR(25), nullable=False)


class ETL_Task_Params(Base):
    __tablename__ = 'etl_task_params'
    __table_args__ = {'quote': False}
    task_id = Column(INTEGER, ForeignKey('etl_tasks.task_id'), primary_key=True, )
    param_name = Column(VARCHAR(400), primary_key=True)
    param_value = Column(BLOB, nullable=True)

    def __init__(self, param_name, param_value):
        self.param_name = param_name
        self.param_value = param_value
        super().__init__()


class ETL_Task_Log(Base):
    __tablename__ = 'etl_task_log'
    __table_args__ = {'quote': False}
    task_id = Column(INTEGER, ForeignKey('etl_tasks.task_id'), primary_key=True, )
    log_entry_ts = Column(DateTime, nullable=True, primary_key=True, )
    log_entry = Column(NCLOB, nullable=True, )


class ETL_Task_Stats(Base):
    __tablename__ = 'etl_task_stats'
    __table_args__ = {'quote': False}
    task_id = Column(INTEGER, ForeignKey('etl_tasks.task_id'), primary_key=True, )
    stat_name = Column(VARCHAR(400), primary_key=True)
    stat_value = Column(VARCHAR(400), nullable=True)

    def __init__(self, stat_name, stat_value):
        self.stat_name = stat_name
        self.stat_value = stat_value
        super().__init__()

    def is_row_count(self):
        return 'rows' in self.stat_name.lower()

    def is_int(self):
        try:
            int(self.stat_value)
            return True
        except ValueError:
            return False


class ETL_Class(Base):
    __tablename__ = 'etl_class'
    __table_args__ = {'quote': False}
    etl_class_id = Column(INTEGER, Sequence('etl_clas_id_seq'), primary_key=True)
    module_name = Column(VARCHAR(400), nullable=False)
    class_name = Column(VARCHAR(400), nullable=True)


class ETL_Class_Dependency(Base):
    __tablename__ = 'etl_class_dependency'
    __table_args__ = {'quote': False}
    etl_class_id = Column(INTEGER, ForeignKey('etl_class.etl_class_id'), primary_key=True, )
    dependent_on_etl_class_id = Column(INTEGER, ForeignKey('etl_class.etl_class_id'), primary_key=True, )

    etl_class = relationship(ETL_Class,
                             foreign_keys=[etl_class_id],
                             backref=backref("dependencies", )
                             )
    dependent_on_etl_task = relationship(ETL_Class,
                                         foreign_keys=[dependent_on_etl_class_id],
                                         backref=backref("depends_on_this", )
                                         )

    def __init__(self, etl_class, dependent_on_etl_task):
        self.etl_class = etl_class
        self.dependent_on_etl_task = dependent_on_etl_task
        super().__init__()


##############################################################################
if __name__ == '__main__':
    config = BI_ETL_Config_Base_From_Ini_Env()
    engine = config.bi_etl.scheduler.db.get_engine()
    Base.metadata.drop_all(engine)
    Base.metadata.create_all(engine)
    print("Model tables created")



"""
Created on Apr 22, 2015

@author: Derek Wood
"""


class ChildRunRequested(object):
    def __init__(self, child_task_id):
        self.child_task_id = child_task_id

    def __repr__(self):
        return "ChildRunRequested(child_task_id={}".format(self.child_task_id)


class ChildRunOK(object):
    def __init__(self, child_task_id):
        self.child_task_id = child_task_id

    def __repr__(self):
        return "ChildRunOK(child_task_id={}".format(self.child_task_id)


class ChildStatusUpdate(object):
    def __init__(self, child_task_id, child_status):
        self.child_task_id = child_task_id
        self.child_status = child_status

    def __repr__(self):
        return "ChildStatusUpdate(child_task_id={}, child_status={}".format(self.child_task_id, self.child_status)


class ChildNotOccupyingCPU(object):
    def __init__(self, child_task_id):
        self.child_task_id = child_task_id

    def __repr__(self):
        return "ChildNotOccupyingCPU(child_task_id={}".format(self.child_task_id)


class ChildOccupyingCPU(object):
    def __init__(self, child_task_id):
        self.child_task_id = child_task_id

    def __repr__(self):
        return "ChildOccupyingCPU(child_task_id={}".format(self.child_task_id)


class SummaryMessage(object):
    def __init__(self, summary_message, append=False):
        self.summary_message = summary_message
        self.append = append

    def __repr__(self):
        return "SummaryMessage(summary_message={}, append={}".format(self.append, self.append)


class ChildSetDisplayName(object):
    def __init__(self, child_task_id, display_name):
        self.child_task_id = child_task_id
        self.display_name = display_name
        
    def __repr__(self):
        return "ChildSetDisplayName(child_task_id={}, display_name={}".format(self.child_task_id, self.display_name)      


"""
Created on Apr 3, 2015

@author: Derek Wood
"""

class NoResultFound(Exception):
    pass

class MultipleResultsFound(Exception):
    pass

class BeforeAllExisting(Exception):    
    def __init__(self, first_existing_row, effective_date):
        self.first_existing_row = first_existing_row
        self.effective_date = effective_date
    
    def __str__(self):
        return 'First row = {} \n not yet started at {}'.format(self.first_existing_row, self.effective_date)

class AfterExisting(Exception):    
    def __init__(self, prior_row, effective_date):
        self.prior_row = prior_row
        self.effective_date = effective_date
        
    def __str__(self):
        return 'Last row = {} \n ends before {}'.format(self.prior_row, self.effective_date)

class ColumnMappingError(Exception):
    def __init__(self, msg):
        self.msg = str(msg)
        
    def __str__(self):
        return self.msg


# -*- coding: utf-8 -*-
"""
Created on Feb 13, 2016

@author: Derek Wood
"""

_frozenset_cache = dict()


def get_cached_frozen_set(set_to_be_frozen):
    if not isinstance(set_to_be_frozen, set):
        set_to_be_frozen = frozenset(set_to_be_frozen)
    
    if set_to_be_frozen in _frozenset_cache:
        return _frozenset_cache[set_to_be_frozen]
    else:
        _frozenset_cache[set_to_be_frozen] = set_to_be_frozen
        return set_to_be_frozen



"""
Created on Nov 17, 2014

@author: Derek Wood
"""
import string
import typing
from datetime import date, timedelta, timezone
from datetime import datetime
from datetime import time
from decimal import Decimal, InvalidOperation
from typing import Union


def strip(s: str):
    """
    Python str.strip() except that it handles None values.
    """
    if s is None:
        return None
    else:
        return s.strip()


def str2int(s: str):
    """
    String to integer
    """
    if s is None or s == '':
        return None
    else:
        return int(s.replace(',', ''))


def int2base(n, base):
    valid_digits = string.digits + string.ascii_uppercase
    if base > len(valid_digits):
        raise ValueError(
            f"int2base requires base <= {len(valid_digits)}. {base} will not work"
        )
    if n < 0:
        sign = -1
        n = abs(n)
    elif n == 0:
        return valid_digits[0]
    else:
        sign = 1

    digits = []
    while n:
        digits.append(valid_digits[n % base])
        n = n // base

    if sign < 0:
        digits.append('-')

    digits.reverse()

    return ''.join(digits)


def str2float(s: str):
    """
    String to floating point
    """
    if s is None or s == '':
        return None
    else:
        try:    
            return float(s.replace(',', ''))
        except ValueError as e:
            if s[-1] in ['-', '+']:
                s2 = s[-1] + s[:-1].replace(',', '')
                return float(s2)
            else:
                raise e


def str2float_end_sign(s: str):
    """
    String to integer
    This version is almost 4 times faster than str2float 
    in handling signs at the end of the string.
    """
    if s is None or s == '':
        return None
    else:
        try:    
            if s[-1] in ['-', '+']:
                s2 = s[-1] + s[:-1].replace(',', '')
                return float(s2)
            else:
                return float(s.replace(',', ''))
        except ValueError:
            return float(s.replace(',', ''))


def str2decimal(s: str):
    """
    String to decimal (AKA numeric)
    """
    if s is None or s == '':
        return None
    else:
        try:  
            s = s.replace(',', '')
            return Decimal(s)
        except InvalidOperation as e:
            if s[-1] in ['-', '+']:
                s2 = s[-1] + s[:-1].replace(',', '')
                return Decimal(s2)
            else:
                raise e


def str2decimal_end_sign(s: str):
    """
    String to decimal (AKA numeric).
    This version is almost 4 times faster than str2decimal
    in handling signs at the end of the string.
    """
    if s is None or s == '':
        return None
    else:
        if s[-1] in ['-', '+']:
            s2 = s[-1] + s[:-1].replace(',', '')
            return Decimal(s2)
        else:
            s = s.replace(',', '')
            return Decimal(s)


def str2date(
        s: str,
        dt_format: str = '%m/%d/%Y',
        ):
    """
    Parse a date (no time) value stored in a string. 
    
    Parameters
    ----------
    s: str
        String value to convert
    dt_format: str
        For format options please see https://docs.python.org/3.5/library/datetime.html#strftime-strptime-behavior
    """
    dt = str2datetime(s, dt_format)
    if dt is not None:
        return date(dt.year, dt.month, dt.day)
    else:
        return None


def str2time(
        s: str,
        dt_format: str = '%H:%M:%S',
        ):
    """
    Parse a time of day value stored in a string. 
    
    Parameters
    ----------
    s: str
        String value to convert
    dt_format: str
        For format options please see
        https://docs.python.org/3/library/datetime.html#strftime-strptime-behavior
    """
    tm = str2datetime(s, dt_format)
    if tm is not None:
        return time(
            tm.hour,
            tm.minute,
            tm.second,
            tm.microsecond,
            tm.tzinfo
            )
    else:
        return None


def str2datetime(
        s: str,
        dt_format: str = '%m/%d/%Y %H:%M:%S',
        ):
    """ 
    Parse a date + time value stored in a string. 
    
    Parameters
    ----------
    s: str
        String value to convert
    dt_format: str
        For format options please see
        https://docs.python.org/3/library/datetime.html#strftime-strptime-behavior
    """
    if s is None or s == '':
        return None
    elif '.%f' in dt_format:
        # Fractional seconds are included in format
        # Try as is and then without in case source drops the Fractional seconds when zero
        try:
            return datetime.strptime(s, dt_format)
        except ValueError as e:
            msg = str(e)
            if 'unconverted data remains' in msg:
                # We might have more digits in the fractional seconds than Python can convert
                msg, remains = msg.split(':')
                remains = remains.strip()
                try:
                    # Make sure what remains is just digits (note this won't work if we have a timezone)
                    int(remains)
                    return datetime.strptime(s[:-1 * len(remains)], dt_format)
                except ValueError:
                    raise e
            else:
                try:
                    return datetime.strptime(s, dt_format.replace('.%f', ''))
                except ValueError:
                    raise e
    # No fractional seconds included in format
    else:
        return datetime.strptime(s, dt_format)


def round_datetime_ms(
        source_datetime: typing.Optional[datetime],
        digits_to_keep: int,
        ):
    """
    Round a datetime value microseconds to a given number of significant digits.
    """
    if source_datetime is None:
        return None
    new_microseconds = round(source_datetime.microsecond, digits_to_keep-6)
    if new_microseconds == 1000000:
        source_datetime = source_datetime.replace(microsecond=0)
        source_datetime += timedelta(seconds=1)
    else:
        source_datetime = source_datetime.replace(microsecond=new_microseconds)
    return source_datetime


def change_tz(
        source_datetime: typing.Optional[datetime],
        from_tzone,
        to_tzone
        ):
    """
    Change time-zones in dates that have no time-zone info, or incorrect time-zone info
    
    Example from_tzone or to_tzone values: ::
        import pytz

        pytz.utc
        pytz.timezone('US/Eastern')
    
    """
    if source_datetime is not None:
        # Apply our source time zone
        result_datetime = source_datetime.replace(tzinfo=from_tzone)
        # Convert to target time zone
        result_datetime = result_datetime.astimezone(to_tzone)
        # Now we strip off the time zone info so it will match what comes out of Oracle
        result_datetime = result_datetime.replace(tzinfo=None)
        return result_datetime    


def get_date_local(dt: datetime) -> datetime:
    if dt.tzinfo is not None:
        local_timezone = datetime.now(timezone.utc).astimezone().tzinfo
        dt = dt.astimezone(local_timezone)
    return dt


def get_date_midnight(dt: datetime) -> datetime:
    dt = get_date_local(dt)
    return dt.replace(hour=0, minute=0, second=0, microsecond=0)


def ensure_datetime(dt: Union[datetime, date]) -> datetime:
    """
    Takes a date or a datetime as input, outputs a datetime
    """
    if isinstance(dt, datetime):
        return dt
    elif isinstance(dt, date):
        return datetime(dt.year, dt.month, dt.day)
    else:
        raise ValueError(f'expected datetime, got {dt}')


def ensure_datetime_dict(
        d: typing.Union[dict, typing.MutableMapping],
        key: str,
        ):
    """
    Takes a dict containing a date or a datetime as input.
    Changes the dict entry to be a datetime
    """
    dt = d[key]
    if isinstance(dt, datetime):
        return
    elif isinstance(dt, date):
        d[key] = datetime(dt.year, dt.month, dt.day)
    else:
        raise ValueError(f'expected datetime for {key}, got {dt}')


def nvl(value, default):
    """
    Pass value through unchanged unless it is NULL (None).
    If it is NULL (None), then return provided default value.
    """
    if (value is None) or (value == ''):
        return default
    else:
        return value


def coalesce(*values):
    for candidate_value in values:
        if candidate_value is not None:
            return candidate_value
    return None


def nullif(v, value_to_null):
    """
    Pass value through unchanged unless it is equal to provided `value_to_null` value. 
    If `v` ==`value_to_null` value then return NULL (None)
    """
    if v == value_to_null:
        return None
    else:
        return v


def default_to_missing(v: str) -> str:
    """
    Same as nvl(v, 'Missing')
    """
    return nvl(v, 'Missing')    


def default_to_invalid(v: str) -> str:
    """
    Same as nvl(v, 'Invalid')
    """
    return nvl(v, 'Invalid')


def default_to_question_mark(v: str) -> str:
    """
    Same as nvl(v, '?')
    """
    return nvl(v, '?')


def default_nines(v: int) -> int:
    """
    Same as nvl(v, -9999)
    """
    return nvl(v, -9999)


def str2bytes_size(str_size: str) -> str:
    """
    Parses a string containing a size in bytes including KB, MB, GB, TB codes
    into an integer with the actual number of bytes (using 1 KB = 1024). 
    """    
    if isinstance(str_size, str):
        str_size = str_size.upper().strip()
        # Trip final B so we can except 10MB or 10M equally
        if str_size[-1] == 'B':
            str_size = str_size[:-1]
            
        # Check for KB
        if str_size[-1] == 'K':
            result = int(str_size[:-1]) * pow(2, 10)
        # Check for MB
        elif str_size[-1] == 'M':
            result = int(str_size[:-1]) * pow(2, 20)
        # Check for GB
        elif str_size[-1] == 'G':
            result = int(str_size[:-1]) * pow(2, 30)
        # Check for TB
        elif str_size[-1] == 'T':
            result = int(str_size[:-1]) * pow(2, 30)
        else:
            result = int(str_size)
    elif str_size is None:
        result = None
    else:
        # return what we were given, just making sure it was an int
        result = int(str_size)
    return result


"""
http://code.activestate.com/recipes/578019/
Bytes-to-human / human-to-bytes converter.
Based on: http://goo.gl/kTQMs
Working with Python 2.x and 3.x.

Author: Giampaolo Rodola' <g.rodola [AT] gmail [DOT] com>
License: MIT
"""

# see: http://goo.gl/kTQMs
SYMBOLS = {
    'customary'     : ['B', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'],
    'customary_ext' : ['byte', 'kilo', 'mega', 'giga', 'tera', 'peta', 'exa',
                       'zetta', 'iotta'],
    'iec'           : ['Bi', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi', 'Yi'],
    'iec_ext'       : ['byte', 'kibi', 'mebi', 'gibi', 'tebi', 'pebi', 'exbi',
                       'zebi', 'yobi'],
}


def bytes2human(
        n: int,
        format_str: str = '%(value).1f %(symbol)s',
        symbols: str = 'customary'
        ) -> str:
    """
    Convert n bytes into a human-readable string based on format_str.
    symbols can be either "customary", "customary_ext", "iec" or "iec_ext",
    see: http://goo.gl/kTQMs

      >>> bytes2human(0)
      '0.0 B'
      >>> bytes2human(0.9)
      '0.0 B'
      >>> bytes2human(1)
      '1.0 B'
      >>> bytes2human(1.9)
      '1.0 B'
      >>> bytes2human(1024)
      '1.0 K'
      >>> bytes2human(1048576)
      '1.0 M'
      >>> bytes2human(1099511627776127398123789121)
      '909.5 Y'

      >>> bytes2human(9856, symbols="customary")
      '9.6 K'
      >>> bytes2human(9856, symbols="customary_ext")
      '9.6 kilo'
      >>> bytes2human(9856, symbols="iec")
      '9.6 Ki'
      >>> bytes2human(9856, symbols="iec_ext")
      '9.6 kibi'

      >>> bytes2human(10000, "%(value).1f %(symbol)s/sec")
      '9.8 K/sec'

      >>> # precision can be adjusted by playing with %f operator
      >>> bytes2human(10000, format_str="%(value).5f %(symbol)s")
      '9.76562 K'
    """
    n = int(n)
    if n < 0:
        raise ValueError("n < 0")
    symbols = SYMBOLS[symbols]
    prefix = {}
    for i, s in enumerate(symbols[1:]):
        prefix[s] = 1 << (i+1)*10
    for symbol in reversed(symbols[1:]):
        if n >= prefix[symbol]:
            value = float(n) / prefix[symbol]
            return format_str % dict(symbol=symbol, value=value)
    return format_str % dict(symbol=symbols[0], value=n)


def human2bytes(s: str) -> int:
    """
    Attempts to guess the string format based on default symbols
    set and return the corresponding bytes as an integer.
    When unable to recognize the format ValueError is raised.

      >>> human2bytes('0 B')
      0
      >>> human2bytes('1 K')
      1024
      >>> human2bytes('1 M')
      1048576
      >>> human2bytes('1 Gi')
      1073741824
      >>> human2bytes('1 tera')
      1099511627776

      >>> human2bytes('0.5kilo')
      512
      >>> human2bytes('0.1  byte')
      0
      >>> human2bytes('1 k')  # k is an alias for K
      1024
      >>> human2bytes('12 foo')
      Traceback (most recent call last):
          ...
      ValueError: can't interpret '12 foo'
    """
    init = s
    num = ""
    while s and s[0:1].isdigit() or s[0:1] == '.':
        num += s[0]
        s = s[1:]
    num = float(num)
    letter = s.strip()
    for _, sset in SYMBOLS.items():
        if letter in sset:
            break
    else:
        if letter == 'k':
            # treat 'k' as an alias for 'K' as per: http://goo.gl/kTQMs
            sset = SYMBOLS['customary']
            letter = letter.upper()
        else:
            raise ValueError("can't interpret %r" % init)
    prefix = {sset[0]: 1}
    for i, s in enumerate(sset[1:]):
        prefix[s] = 1 << (i+1)*10
    return int(num * prefix[letter])


def replace_tilda(e):
    """
    Used for unicode error to replace invalid ascii with ~

    Apply this with this code

    .. code-block:: python

        codecs.register_error('replace_tilda', replace_tilda)
        ...
        bytes_value = str_value.encode('ascii', errors='replace_tilda')

    See https://docs.python.org/3/library/codecs.html#codecs.register_error
    """
    return u'~', e.start + 1



"""
Created on Nov 18, 2014

@author: Derek Wood
"""
import logging
import math
import typing


def dict_to_list(obj,
                 depth=0,
                 **kwargs
                 ):
    # This proc will catch all exceptions in many places.
    # This is done because dict_to_list is used to print debug
    # and error messages and shouldn't raise exceptions itself. 
    # pylint: disable= broad-except

    # Optional Arguments
    indent_per_level = kwargs.get('indent_per_level', 2)
    depth_limit = kwargs.get('depth_limit')
    entry_name = kwargs.get('entry_name')
    if entry_name is not None and not isinstance(entry_name, str):
        entry_name = str(entry_name)
    # Don't pass entry name to next call
    if 'entry_name' in kwargs:
        del kwargs['entry_name']
    item_limit = kwargs.get('item_limit')
    sorted_dicts = kwargs.get('sorted_dicts')
    show_type = kwargs.get('show_type', True)
    show_length = kwargs.get('show_length', True)
    show_list_item_number = kwargs.get('show_list_item_number', True)
    type_formats = kwargs.get('type_formats')
    display_internal_dict = kwargs.get('display_internal_dict', False)
    result = []
    name_value_separator = ': '
    # Performance tweak.  Store result.append reference once locally since that's faster
    result_add = result.append
    try:

        if indent_per_level * depth > 0:
            indent = (' ' * indent_per_level * depth)
        else:
            indent = ''
        if entry_name is not None:
            entry_header = [entry_name, name_value_separator]
        else:
            entry_header = []
        if show_length and hasattr(obj, '__len__'):
            try:
                entry_header.append("length = " + str(len(obj)) + " ")
            except TypeError:
                pass

        if depth_limit:
            if depth >= depth_limit:
                result_add(indent)
                return result + ["--Depth Limit--"]
        item_number = 0
        # If we have a format specified for this type
        if type_formats is not None and type(obj) in type_formats:
            result_add(indent)
            result += entry_header
            try:
                line_format = '{:' + type_formats[type(obj)] + '}'
                result_add(line_format.format(obj))
            except Exception as e:
                result_add(repr(e))
            if show_type:
                result_add(' ' + str(type(obj)))
        elif isinstance(obj, int):
            result_add(indent)
            result += entry_header
            try:
                result_add(str(obj))
            except Exception as e:
                result_add(repr(e))
            if show_type:
                result_add(" <type 'int'>")
        elif isinstance(obj, str):
            result_add(indent)
            result += entry_header
            try:
                result_add(obj)
            except Exception as e:
                result_add(repr(e))
            if show_type:
                result_add(' ')
                result_add(str(type(obj)))
        elif isinstance(obj, bytes):
            result_add(indent)
            result += entry_header
            try:
                result_add(str(obj))
            except Exception as e:
                result_add(repr(e))
            if show_type:
                result_add(' ')
                result_add(str(type(obj)))
        elif isinstance(obj, type):
            result_add(indent)
            result += entry_header
            try:
                result_add(str(obj))
            except Exception as e:
                result_add(repr(e))
            if show_type:
                result_add(' ')
                result_add(str(type(obj)))
        # -----------------------------------------------------------------------
        # Row like (use columns_in_order to get the columns in the correct order)
        elif hasattr(obj, 'columns_in_order'):
            if show_type:
                if len(entry_header) != 0:
                    entry_header.append(' ')
                entry_header.append(str(type(obj)))
            if len(entry_header) != 0:
                result_add(indent)
                result += entry_header
                result_add('\n')
            for k in obj.columns_in_order:
                v = obj[k]
                item_number += 1
                if item_number > 1:
                    result += ['\n']
                if item_limit:
                    if item_number >= item_limit:
                        result_add("--Item Limit--")
                        break
                result += dict_to_list(v, depth=depth + 1, entry_name=k, **kwargs)
        # -----------------------------------------------------------------------
        # Dict like object
        elif hasattr(obj, 'items'):
            if show_type:
                if len(entry_header) != 0:
                    entry_header.append(' ')
                entry_header.append(str(type(obj)))
            if len(entry_header) != 0:
                result_add(indent)
                result += entry_header
                result_add('\n')
            try:
                if sorted_dicts:
                    items = sorted(obj.items())
                else:
                    items = list(obj.items())
            except TypeError as e:
                result_add(repr(e))
                return result

            for k, v in items:
                item_number += 1
                if item_number > 1:
                    result += ['\n']
                if item_limit:
                    if item_number >= item_limit:
                        result_add("--Item Limit--")
                        break
                result += dict_to_list(v, depth=depth + 1, entry_name=k, **kwargs)
        # -----------------------------------------------------------------------
        # list like
        elif hasattr(obj, '__iter__'):
            if show_type:
                if len(entry_header) != 0:
                    entry_header.append(' ')
                entry_header.append(str(type(obj)))
            if len(entry_header) != 0:
                result_add(indent)
                result += entry_header
                result_add('\n')
            try:
                for v in obj:
                    item_number += 1
                    if item_number > 1:
                        result_add('\n')
                    if item_limit:
                        if item_number >= item_limit:
                            result_add("--Item Limit--")
                            break
                    if show_list_item_number:
                        kwargs['entry_name'] = ('list item ' + str(item_number))
                    result += dict_to_list(v, depth=depth + 1, **kwargs)
            except IOError as e:
                result_add(repr(e))
            except TypeError as e:
                result_add(repr(e))
        # -----------------------------------------------------------------------
        # SQLAlchemy ORM object
        elif hasattr(obj, '__table__'):
            if show_type:
                if len(entry_header) != 0:
                    entry_header.append(' ')
                entry_header.append(str(type(obj)))
            if len(entry_header) != 0:
                result_add(indent)
                result += entry_header
                result_add('\n')
            for column in obj.__table__.columns:
                attr = column.name
                result_add('\n')
                v = getattr(obj, attr)
                result += dict_to_list(v, depth=depth + 1, entry_name=attr, **kwargs)
                if show_type:
                    result_add(' ' + str(type(v)))
        elif display_internal_dict and hasattr(obj, '__dict__'):
            result_add(indent)
            result += entry_header
            result_add(str(obj))
            if show_type:
                try:
                    result_add(' ' + str(type(obj)))
                except Exception as e:
                    result_add(repr(e))
            result_add('\n')
            result += dict_to_list(obj.__dict__, depth=depth + 1, entry_name=entry_name + '.__dict__', **kwargs)
        # -----------------------------------------------------------------------
        # Otherwise use the default str
        else:
            result_add(indent)
            result += entry_header
            result_add(str(obj))
            if show_type:
                try:
                    result_add(' ' + str(type(obj)))
                except Exception as e:
                    result_add(repr(e))
        return result
    except Exception as e:
        # print((traceback.format_exc()))
        return [repr(e)]


def dict_to_str(obj,
                depth=0,
                **kwargs
                ):
    """
    Parameters:
    obj is the object to convert to a string format

    entry_name is the main title to put at the top (default blank)

    depth is the starting depth (default 0)

    indent_per_level is the number of spaces to indent per depth level (default 2)

    depth_limit is the limit on how many levels deep to recurse (default no limit)

    item_limit is the limit on how many items in a given container to output (default no limit)

    show_type is a boolean indicating if the type of each entry should be included (default True)

    show_list_item_number is a boolean indicating if the sequence number should be included for list entries
    (default True)

    type_formats is a dictionary mapping types to print format specifiers

    """
    return ''.join(dict_to_list(obj, depth, **kwargs))


def dict_to_pairs(obj, prefix=None, delimit='.'):
    for k in list(obj.keys()):
        val = obj[k]

        if prefix is None:
            name = k
        else:
            name = prefix + delimit + k

        if hasattr(val, 'keys') and hasattr(val, '__getitem__'):
            for stat_tuple in dict_to_pairs(val, prefix=name, delimit=delimit):
                yield stat_tuple
        else:
            yield (name, obj[k])


def log_logging_level(log):
    original_level = log.getEffectiveLevel()
    if original_level > logging.INFO:
        # If we aren't at INFO level or more detailed, then temporarily set it to INFO and restore it later
        log.setLevel(logging.INFO)
    log.info('Logging level is {name} ({num})'.format(
        num=log.getEffectiveLevel(),
        name=logging.getLevelName(original_level))
    )
    if original_level > logging.INFO:
        # Restore logging level
        log.setLevel(original_level)


def get_integer_places(number):
    abs_number = int(abs(number))
    if abs_number == 0:
        return 0
    # For numbers larger than 999999999999997 the log10 function's rounding causes the result to be too big
    elif abs_number <= 999999999999997:
        return int(math.log10(abs_number)) + 1
    else:
        # str tested as faster than trying to count digits via exponentiation
        return len(str(abs_number))


def single_quote_value(value) -> str:
    return f"'{value}'"


def double_quote_value(value) -> str:
    return f'"{value}"'


def quote_delimited_list(
        value_list,
        delimiter=', ',
        quote_function: typing.Callable[[object], str] = single_quote_value,
        ) -> str:
    return delimiter.join([quote_function(v) for v in value_list])
